# -*- coding: utf-8 -*-
from copy import copy
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import List
from typing import Optional

import matplotlib as mpl

# from nata.plots import Figure
from nata.plots.data import PlotData
from nata.plots.types import PlotTypes


@dataclass
class Axes:
    """Container of parameters and parent and child objects (including\
    plotting backend-related objects) relevant to draw a figure axes.

    Parameters
    ----------
    xlim: ``tuple``, optional
        Limits of the horizontal axis in the format ``(min,max)``. If not
        provided, it is inferred from the dataset(s) represented in the axes.

    ylim: ``tuple``, optional
        Same as ``xlim`` for the vertical axis.

    xlabel: ``str``, optional
        Label of the horizontal axis. If not provided, it is inferred from the
        dataset(s) represented in the axis.

    ylabel: ``str``, optional
        Same as ``xlabel`` for the vertical axis.

    xscale: ``{'linear','log', 'symlog'}``, optional
        Scale of the horizontal axis. If not provided, defaults to
        ``'linear'``. If set to ``'symlog'``, ``linthreshx`` is required.

    yscale: ``{'linear','log', 'symlog'}``, optional
        Same as ``xscale`` for the vertical axis.

    linthreshx: ``float``, optional
        Range within which the horizontal axis is linear. Applicable only when
        ``xscale`` is set to ``'symlog'``.

    linthreshy: ``float``, optional
        Same as ``linthreshx`` for the vertical axis.

    title: ``str``, optional
        Axes title. If not provided, it is inferred from the dataset(s)
        represented in the axes.

    legend_show: ``bool``, optional
        Controls the visibility of the axes legend, when applicable. If not
        provided, defaults to ``True``.

    legend_loc: ``str``, optional
        Controls the position of the axes legend, when applicable. See
        :meth:`matplotlib.axes.Axes.legend` for available options. If not
        provided, defaults to ``'upper right'``.

    legend_frameon: ``bool``, optional
        Controls the visibility of the axes legend frame. If not provided,
        defaults to ``False``.

    cb_show: ``bool``, optional
        Controls the visibility of the axes colorbar, when applicable. If not
        provided, defaults to ``True``.

    index: ``int``, optional
        Position of the axes in the parent figure. Must be between ``0`` and
        ``N-1``, where ``N`` is the number of child axes objects in the parent
        figure. Increases along rows before columns.

    """

    # style properties
    xlim: Optional[tuple] = None
    ylim: Optional[tuple] = None
    xlabel: Optional[str] = None
    ylabel: Optional[str] = None
    title: Optional[str] = None
    xscale: Optional[str] = "linear"
    yscale: Optional[str] = "linear"
    linthreshx: Optional[float] = None
    linthreshy: Optional[float] = None
    aspect: Optional[str] = "auto"
    legend_show: Optional[bool] = True
    legend_loc: Optional[str] = "upper right"
    legend_frameon: Optional[bool] = False
    cb_show: Optional[bool] = True

    # other parameters
    index: Optional[int] = 0

    # flags for automatic style properties
    xlim_auto: bool = field(init=False, default=None)
    ylim_auto: bool = field(init=False, default=None)
    xlabel_auto: bool = field(init=False, default=None)
    ylabel_auto: bool = field(init=False, default=None)
    title_auto: bool = field(init=False, default=None)

    # backend objects
    ax: Any = field(init=False, repr=False, default=None)
    cb: Any = field(init=False, repr=False, default=None)
    legend: Any = field(init=False, repr=False, default=None)

    # parent figure object
    fig: Any = None

    # child plot objects
    plots: List[PlotTypes] = field(init=False, repr=False, default_factory=list)

    def __post_init__(self):
        # initialize axes backend
        self.init_backend()

    # backend methods
    # TODO: generalize the following methods for arbitrary backend
    def init_backend(self):
        with mpl.rc_context(fname=self.fig.fname, rc=self.fig.rc):
            self.ax = self.fig.fig.add_subplot(
                self.fig.nrows, self.fig.ncols, self.index
            )

    def clear_backend(self):
        for plot in self.plots:
            plot.clear()

        self.clear_colorbar()
        self.clear_legend()

        self.ax.clear()
        self.ax.remove()
        self.ax = None

    def reset_backend(self):
        self.clear_backend()
        self.init_backend()

    def update_backend(self):
        with mpl.rc_context(fname=self.fig.fname, rc=self.fig.rc):
            ax = self.ax

            if self.xscale == "symlog" and self.linthreshx is not None:
                ax.set_xscale(self.xscale, linthreshx=self.linthreshx)
            else:
                ax.set_xscale(self.xscale)

            if self.yscale == "symlog" and self.linthreshy is not None:
                ax.set_yscale(self.yscale, linthreshy=self.linthreshy)
            else:
                ax.set_yscale(self.yscale)

            if self.xlim[0] != self.xlim[1]:
                ax.set_xlim(self.xlim)

            if self.ylim[0] != self.ylim[1]:
                ax.set_ylim(self.ylim)

            # set axes labels
            ax.set_xlabel(self.xlabel)
            ax.set_ylabel(self.ylabel)

            # set title
            ax.set_title(label=self.title)

            # set aspect ratio
            ax.set_aspect(self.aspect)

    def add_plot(
        self,
        plot_type: Optional[PlotTypes] = None,
        plot: Optional[PlotTypes] = None,
        data: Optional[PlotData] = None,
        style: Optional[dict] = dict(),
    ):
        if plot is None:
            plot = plot_type(axes=self, data=data, **style)
        else:
            plot.axes = self

        self.plots.append(plot)

        return plot

    def update_plot_options(self):
        if self.xlim_auto:
            self.xlim = (
                min([p._default_xlim()[0] for p in self.plots]),
                max([p._default_xlim()[1] for p in self.plots]),
            )

        if self.ylim_auto:
            self.ylim = (
                min([p._default_ylim()[0] for p in self.plots]),
                max([p._default_ylim()[1] for p in self.plots]),
            )

        if self.xlabel_auto:
            units = [p._xunits() for p in self.plots]
            xlabels = [p._default_xlabel(units=False) for p in self.plots]

            if len(set(units)) == 1 and len(set(xlabels)) == 1:
                self.xlabel = set(xlabels).pop()
                if set(units).pop():
                    self.xlabel += " [" + set(units).pop() + "]"
            elif len(set(units)) == 1:
                xlabels = [p._default_xlabel(units=False) for p in self.plots]
                self.xlabel = ", ".join(xlabels)
                if set(units).pop():
                    self.xlabel += " [" + set(units).pop() + "]"
            else:
                xlabels = [p._default_xlabel(units=True) for p in self.plots]
                self.xlabel = ", ".join(xlabels)

        if self.ylabel_auto:
            units = [p._yunits() for p in self.plots]
            ylabels = [p._default_ylabel(units=False) for p in self.plots]

            if len(set(units)) == 1 and len(set(ylabels)) == 1:
                self.ylabel = set(ylabels).pop()
                if set(units).pop():
                    self.ylabel += " [" + set(units).pop() + "]"
            elif len(set(units)) == 1:
                ylabels = [p._default_ylabel(units=False) for p in self.plots]
                self.ylabel = ", ".join(ylabels)
                if set(units).pop():
                    self.ylabel += " [" + set(units).pop() + "]"
            else:
                ylabels = [p._default_ylabel(units=True) for p in self.plots]
                self.ylabel = ", ".join(ylabels)

        if self.title_auto:
            titles = [p._default_title() for p in self.plots]

            if len(set(titles)) == 1:
                self.title = set(titles).pop()
            else:
                self.title = None

    def update(self):
        self.update_plot_options()
        self.update_backend()

        if len(self.plots) > 1:
            self.init_legend()

    def redo_plots(self):

        self.reset_backend()

        for plot in self.plots:
            plot.build_canvas()

        self.update()

    # legend methods
    def init_legend(self):
        if self.legend_show:
            handles, labels = self.ax.get_legend_handles_labels()
            with mpl.rc_context(fname=self.fig.fname, rc=self.fig.rc):
                # show legend
                self.legend = self.ax.legend(
                    handles=handles,
                    labels=labels,
                    loc=self.legend_loc,
                    frameon=self.legend_frameon,
                )

    def clear_legend(self):
        if self.legend:
            self.legend.remove()
            self.legend = None

    # colorbar methods
    def init_colorbar(self, plot: PlotTypes):
        if self.cb_show:
            with mpl.rc_context(fname=self.fig.fname, rc=self.fig.rc):
                # show colorbar
                self.cb = self.ax.get_figure().colorbar(plot.h, aspect=30)

                # set colorbar title)
                self.cb.set_label(label=plot.cb_title)

    def clear_colorbar(self):
        if self.cb:
            self.cb.remove()
            self.cb = None

    def copy(self):
        new = copy(self)

        for plot in new.plots:
            plot.axes = new

        return new

    @classmethod
    def style_attrs(self) -> List[str]:
        return [
            "xlim",
            "ylim",
            "xlabel",
            "ylabel",
            "title",
            "xscale",
            "yscale",
            "linthreshx",
            "linthreshy",
            "aspect",
            "legend_show",
            "legend_loc",
            "legend_frameon",
            "cb_show",
        ]



# -*- coding: utf-8 -*-
from typing import Optional

import numpy as np
import numpy.fft as fft

from nata.axes import GridAxis
from nata.containers import GridDataset
from nata.plugins.register import register_container_plugin


@register_container_plugin(GridDataset, name="fft")
def fft_grid_dataset(
    dataset: GridDataset, type: Optional[str] = "abs",
) -> GridDataset:
    """Computes the Fast Fourier Transform (FFT) of a single/multiple\
       iteration :class:`nata.containers.GridDataset` along all grid axes\
       using `numpy's fft module`__.

        .. _fft: https://numpy.org/doc/stable/reference/routines.fft.html
        __ fft_

        Parameters
        ----------
        type: ``{'abs', 'real', 'imag', 'full'}``, optional
            Defines the component of the FFT selected for output. Available
            values are  ``'abs'`` (default), ``'real'``, ``'imag'`` and
            ``'full'``, which correspond to the absolute value, real component,
            imaginary component and full (complex) result of the FFT,
            respectively.

        Returns
        ------
        :class:`nata.containers.GridDataset`:
            Selected FFT component along all grid axes of ``dataset``.

        Examples
        --------
        To obtain the FFT of a :class:`nata.containers.GridDataset`, a simple
        call to the ``fft()`` method is enough. In the following example, we
        compute the FFT of a one-dimensional
        :class:`nata.containers.GridDataset`.

        >>> from nata.containers import GridDataset
        >>> import numpy as np
        >>> x = np.linspace(100)
        >>> arr = np.exp(-(x-len(x)/2)**2)
        >>> ds = GridDataset(arr[np.newaxis])
        >>> ds_fft = ds.fft()

    """

    fft_data = np.array(dataset)
    fft_axes = np.arange(len(dataset.grid_shape)) + 1

    fft_data = fft.fftn(
        fft_data if len(dataset) > 1 else fft_data[np.newaxis], axes=fft_axes
    )
    fft_data = fft.fftshift(fft_data, axes=fft_axes)

    if type == "real":
        fft_data = np.real(fft_data)
        label = f"Re(FFT({dataset.label}))"
        name = f"fftr_{dataset.name}"
    elif type == "imag":
        fft_data = np.imag(fft_data)
        label = f"Im(FFT({dataset.label}))"
        name = f"ffti_{dataset.name}"
    elif type == "abs":
        fft_data = np.abs(fft_data)
        label = f"|FFT({dataset.label})|"
        name = f"ffta_{dataset.name}"
    else:
        label = f"FFT({dataset.label})"
        name = f"fft_{dataset.name}"

    axes = []
    for a in dataset.axes["grid_axes"]:
        delta = [
            (np.max(a_ts) - np.min(a_ts)) / len(np.array(a_ts)) / (2.0 * np.pi)
            for a_ts in a
        ]
        axis_data = fft.fftshift(
            [
                fft.fftfreq(len(np.array(a_ts)), delta[idx])
                for idx, a_ts in enumerate(a)
            ],
            axes=-1,
        )
        axes.append(
            GridAxis(
                axis_data,
                name=f"k_{a.name}",
                label=f"k_{{{a.label}}}",
                unit=f"1/({a.unit})",
            )
        )

    return GridDataset(
        fft_data,
        name=name,
        label=label,
        unit=dataset.unit,
        grid_axes=axes,
        time=dataset.axes["time"],
        iteration=dataset.axes["iteration"],
    )



# -*- coding: utf-8 -*-
from copy import deepcopy
from typing import List

import numpy as np

from nata.containers import ParticleDataset
from nata.containers import ParticleQuantity
from nata.plugins.register import register_container_plugin


@register_container_plugin(ParticleDataset, name="filter")
def filter_particle_dataset(
    dataset: ParticleDataset,
    mask: List[bool] = None,
    quantities: List[str] = None,
    slicing: slice = None,
) -> ParticleDataset:
    """Filters a :class:`nata.containers.ParticleDataset` according to a\
       selection of quantities.

        Parameters
        ----------
        mask: ``np.ndarray``, optional
            Array of booleans indicating the particles to be filtered. Particles
            with ``True`` (``False``) mask entries are selected (hidden). The
            shape of ``mask`` must match that of each particle quantity.

        slicing: ``slice``, optional
            Slice of particles to be filtered. Acts only on particle indices
            and not on time, as time slicing should be done on the dataset.
            When provided together with the ``mask argument``, slicing is done
            on the masked array.

        quantities: ``list``, optional
            List of quantities to be filtered, ordered by the way they should be
            sorted in the returned dataset.

        Returns
        ------
        :class:`nata.containers.ParticleDataset`:
            Filtered dataset with only the quantities selected in
            ``quantities``.

        Examples
        --------
        The filter plugin is used to get dataset with only a selection, say
        ``'x1'`` and ``'x2'``, of its quantities.

        >>> from nata.containers import ParticleDataset
        >>> ds = ParticleDataset("path/to"file")
        >>> ds_flt = ds.filter(quantities=["x1","p1"])

    """

    dataset_c = deepcopy(dataset)

    if quantities is not None:
        if not isinstance(quantities, list) or not all(
            isinstance(quant, str) for quant in quantities
        ):
            raise TypeError(f"'quantities' must be a list of strings.")

        quants = {
            quant: dataset_c.quantities[quant]
            for quant in quantities
            if quant in dataset_c.quantities.keys()
        }

        if not bool(quants):
            raise ValueError(
                f"None of the quantities in 'quantities' were found in the "
                + "dataset."
            )

    else:
        quants = dataset_c.quantities

    if mask is not None:
        if not isinstance(mask, np.ndarray) or not mask.dtype == bool:
            raise TypeError(f"'mask' must be an 'np.ndarray' of dtype 'bool'.")

        quant = next(iter(quants.values()))
        if quant.data.shape != mask.shape:
            raise ValueError(
                f"'mask.shape' ({mask.shape}) does not match the shape of "
                + "each quantity ({quant.data.shape})."
            )

        for name, quant in quants.items():
            quants[name].data[~mask] = np.ma.masked

    if slicing is not None:
        if not isinstance(slicing, slice):
            raise TypeError(f"'slicing' must be of type 'slice'.")

        if len(dataset_c) > 1:
            slicing = (slice(None), slicing)

        quants = {
            quant.name: ParticleQuantity(
                data=quant.data[slicing],
                name=quant.name,
                label=quant.label,
                unit=quant.unit,
                dtype=quant.data.dtype,
            )
            for quant in quants.values()
        }

    return ParticleDataset(
        iteration=dataset_c.axes["iteration"],
        time=dataset_c.axes["time"],
        name=dataset_c.name,
        quantities=quants,
    )



# -*- coding: utf-8 -*-
import sys

# cached_property was incorporated in  python 3.8+
if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    cached_property = property

    # something is wrong in the implementation
    # class cached_property:
    #     def __init__(self, func):
    #         self._func = func

    #     def __get__(self, instance, owner=None):
    #         assert instance is not None
    #         ret = \
    #           instance.__dict__[self._func.__name__] = self._func(instance)
    #         return ret



# -*- coding: utf-8 -*-
from typing import Optional
from typing import Union
from warnings import warn

import numpy as np

from nata.containers import GridDataset
from nata.plots.axes import Axes
from nata.plots.data import PlotData
from nata.plots.data import PlotDataAxis
from nata.plots.figure import Figure
from nata.plots.helpers import filter_style
from nata.plots.plans import AxesPlan
from nata.plots.plans import FigurePlan
from nata.plots.plans import PlotPlan
from nata.plots.types import DefaultGridPlotTypes
from nata.plugins.register import register_container_plugin
from nata.utils.env import inside_notebook


@register_container_plugin(GridDataset, name="plot_data")
def grid_plot_data(dataset: GridDataset) -> PlotData:

    a = []

    for ds_a in dataset.axes["grid_axes"]:
        new_a = PlotDataAxis(
            name=ds_a.name,
            label=ds_a.label,
            units=ds_a.unit,
            data=np.array(ds_a),
        )

        a.append(new_a)

    d = PlotData(
        name=dataset.name,
        label=dataset.label,
        units=dataset.unit,
        data=np.array(dataset),
        time=np.array(dataset.axes["time"]),
        time_units=dataset.axes["time"].unit,
        axes=a,
    )

    return d


@register_container_plugin(GridDataset, name="plot_type")
def grid_plot_type(dataset: GridDataset) -> PlotData:
    return DefaultGridPlotTypes[len(dataset.grid_shape)]


@register_container_plugin(GridDataset, name="plot")
def plot_grid_dataset(
    dataset: GridDataset,
    fig: Optional[Figure] = None,
    axes: Optional[Axes] = None,
    style: Optional[dict] = {},
    interactive: Optional[bool] = True,
    n: Optional[int] = 0,
) -> Union[Figure, None]:
    """Plots a single/multiple iteration :class:`nata.containers.GridDataset`\
       using a :class:`nata.plots.types.LinePlot` or\
       :class:`nata.plots.types.ColorPlot` if the dataset is one- or\
       two-dimensional, respectively.

        Parameters
        ----------
        fig: :class:`nata.plots.Figure`, optional
            If provided, the plot is drawn on ``fig``. The plot is drawn on
            ``axes`` if it is a child axes of ``fig``, otherwise a new axes
            is created on ``fig``. If ``fig`` is not provided, a new
            :class:`nata.plots.Figure` is created.

        axes: :class:`nata.plots.Axes`, optional
            If provided, the plot is drawn on ``axes``, which must be an axes
            of ``fig``. If ``axes`` is not provided or is provided without a
            corresponding ``fig``, a new :class:`nata.plots.Axes` is created in
            a new :class:`nata.plots.Figure`.

        style: ``dict``, optional
            Dictionary that takes a mix of style properties of
            :class:`nata.plots.Figure`, :class:`nata.plots.Axes` and any plot
            type (see :class:`nata.plots.types.LinePlot` or
            :class:`nata.plots.types.ColorPlot`).

        interactive: ``bool``, optional
            Controls wether interactive widgets should be shown with the plot
            to allow for temporal navigation. Only applicable if ``dataset``
            has multiple iterations.

        n: ``int``, optional
            Selects the index of the iteration to be shown initially. Only
            applicable if ``dataset`` has multiple iterations, .

        Returns
        ------
        :class:`nata.plots.Figure` or ``None``:
            Figure with plot built based on ``dataset``. Interactive widgets
            are shown with the figure if ``dataset`` has multiple iterations,
            in which case this method returns  ``None``.

        Examples
        --------
        To get a plot with default style properties in a new figure, simply
        call the ``.plot()`` method of the dataset.

        >>> from nata.containers import GridDataset
        >>> import numpy as np
        >>> arr = np.arange(10)
        >>> ds = GridDataset.from_array(arr)
        >>> fig = ds.plot()

        In case a :class:`nata.plots.Figure` is returned by the method, it can
        be shown by calling the :func:`nata.plots.Figure.show` method.

        >>> fig.show()

        To draw a new plot on ``fig``, we can pass it as an argument to the
        ``.plot()`` method. If ``axes`` is provided, the new plot is drawn on
        the selected axes.

        >>> ds2 = GridDataset.from_array(arr**2)
        >>> fig = ds2.plot(fig=fig, axes=fig.axes[0])


    """

    p_plan = PlotPlan(
        dataset=dataset, style=filter_style(dataset.plot_type(), style)
    )

    a_plan = AxesPlan(
        axes=axes, plots=[p_plan], style=filter_style(Axes, style)
    )

    f_plan = FigurePlan(
        fig=fig, axes=[a_plan], style=filter_style(Figure, style)
    )

    if len(dataset) > 1:
        if inside_notebook():
            if interactive:
                f_plan.build_interactive(n)
            else:
                return f_plan[n].build()
        else:
            # TODO: remove last line from warn
            warn(
                f"Plotting only iteration with index n={str(n)}."
                + " Interactive plots of multiple iteration datasets are not"
                + " supported outside notebook environments."
            )
            return f_plan[n].build()

    else:
        return f_plan.build()



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import List
from typing import Optional

import matplotlib.colors as clr
import numpy as np

from nata.plots.types import BasePlot


@dataclass
class ColorPlot(BasePlot):
    """Color plot class.

    Parameters
    ----------
    vmin: ``float``, optional
        Minimum of the colorbar axis. If not  provided, it is inferred from
        the dataset represented in the plot.

    vmax: ``float``, optional
        Same as ``vmin`` for the maximum of the colorbar axis.

    cb_title: ``str``, optional
        Colorbar title. If not provided, it is inferred from the dataset
        represented in the plot.

    cb_scale: ``{'linear','log', 'symlog'}``, optional
        Scale of the colorbar. If not provided, defaults to ``'linear'``.

    cb_map: ``str``, optional
        Colormap used to represent the data. See
        :func:`matplotlib.pyplot.colormaps` for available options. If not
        provided, defaults to ``rainbow``.

    cb_linthresh: ``float``, optional
        Range within which the colorbar axis is linear. Applicable only when
        ``cb_scale`` is set to ``'symlog'``. If not provided, defaults to
        ``1e-5``.

    interpolation: ``str``, optional
        Interpolation method used. See :func:`matplotlib.pyplot.imshow` for
        available options. If not provided, defaults to ``none``.

    """

    vmin: Optional[float] = None
    vmax: Optional[float] = None
    cb_map: Optional[str] = "rainbow"
    cb_scale: Optional[str] = "linear"
    cb_linthresh: Optional[float] = 1e-5
    cb_title: Optional[str] = None
    interpolation: Optional[str] = "none"

    def __post_init__(self):
        if self.cb_title is None:
            self.cb_title = self.data.get_label()

        super().__post_init__()

    def _default_xlim(self):
        return (self.data.axes[0].min, self.data.axes[0].max)

    def _default_ylim(self):
        return (self.data.axes[1].min, self.data.axes[1].max)

    def _default_xlabel(self, units=True):
        return self.data.axes[0].get_label(units)

    def _default_ylabel(self, units=True):
        return self.data.axes[1].get_label(units)

    def _default_title(self):
        return self.data.get_time_label()

    def _default_label(self):
        return self.data.get_label(units=False)

    def _xunits(self):
        return f"${self.data.axes[0].units}$" if self.data.axes[0].units else ""

    def _yunits(self):
        return f"${self.data.axes[1].units}$" if self.data.axes[1].units else ""

    def build_canvas(self):
        # get plot extent and data
        extent = (
            self.data.axes[0].min,
            self.data.axes[0].max,
            self.data.axes[1].min,
            self.data.axes[1].max,
        )
        z = np.transpose(self.data.data)

        # build color map norm
        if self.cb_scale == "log":
            # convert values to positive numbers
            z = np.abs(z) + 1e-16

            # adjust min and max values
            # TODO: do this only if vmin was not init
            # self.vmin = np.min(z)

            # if self.vmax_auto:
            # self.vmax = np.max(z)

            # set color map norm
            self.cb_norm = clr.LogNorm(vmin=self.vmin, vmax=self.vmax)
        elif self.cb_scale == "symlog":
            # set color map norm
            self.cb_norm = clr.SymLogNorm(
                vmin=self.vmin,
                vmax=self.vmax,
                linthresh=self.cb_linthresh,
                base=10,
            )
        else:
            self.cb_norm = clr.Normalize(vmin=self.vmin, vmax=self.vmax)

        # build plot
        self.h = self.axes.ax.imshow(
            z,
            extent=extent,
            origin="lower",
            cmap=self.cb_map,
            norm=self.cb_norm,
            interpolation=self.interpolation,
        )

        self.axes.init_colorbar(plot=self)

    @classmethod
    def style_attrs(cls) -> List[str]:
        return [
            "vmin",
            "vmax",
            "cb_map",
            "cb_scale",
            "cb_linthresh",
            "cb_title",
            "interpolation",
        ] + BasePlot.style_attrs()



# -*- coding: utf-8 -*-
from typing import Optional
from warnings import warn

import numpy as np

from nata.containers import ParticleDataset
from nata.plots.axes import Axes
from nata.plots.data import PlotData
from nata.plots.data import PlotDataAxis
from nata.plots.figure import Figure
from nata.plots.helpers import filter_style
from nata.plots.plans import AxesPlan
from nata.plots.plans import FigurePlan
from nata.plots.plans import PlotPlan
from nata.plots.types import DefaultParticlePlotType
from nata.plugins.register import register_container_plugin
from nata.utils.env import inside_notebook


@register_container_plugin(ParticleDataset, name="plot_data")
def particle_plot_data(dataset: ParticleDataset) -> PlotData:
    a = []
    d = []

    for quant in dataset.quantities.values():
        new_a = PlotDataAxis(
            name=quant.name, label=quant.label, units=quant.unit
        )

        a.append(new_a)
        d.append(quant.data)

    return PlotData(
        name=dataset.name,
        label=dataset.name,
        units="",
        data=d,
        time=np.array(dataset.axes["time"]),
        time_units=dataset.axes["time"].unit,
        axes=a,
    )


@register_container_plugin(ParticleDataset, name="plot_type")
def particle_plot_type(dataset: ParticleDataset) -> PlotData:
    return DefaultParticlePlotType


@register_container_plugin(ParticleDataset, name="plot")
def plot_particle_dataset(
    dataset: ParticleDataset,
    fig: Optional[Figure] = None,
    axes: Optional[Axes] = None,
    style: dict = dict(),
    interactive: bool = True,
    n: int = 0,
):
    """Plots a single/multiple iteration :class:`nata.containers.ParticleDataset`\
       using a :class:`nata.plots.types.ScatterPlot`.

        Parameters
        ----------
        fig: :class:`nata.plots.Figure`, optional
            If provided, the plot is drawn on ``fig``. The plot is drawn on
            ``axes`` if it is a child axes of ``fig``, otherwise a new axes
            is created on ``fig``. If ``fig`` is not provided, a new
            :class:`nata.plots.Figure` is created.

        axes: :class:`nata.plots.Axes`, optional
            If provided, the plot is drawn on ``axes``, which must be an axes
            of ``fig``. If ``axes`` is not provided or is provided without a
            corresponding ``fig``, a new :class:`nata.plots.Axes` is created in
            a new :class:`nata.plots.Figure`.

        style: ``dict``, optional
            Dictionary that takes a mix of style properties of
            :class:`nata.plots.Figure`, :class:`nata.plots.Axes` and any plot
            type (see :class:`nata.plots.types.ScatterPlot`).

        interactive: ``bool``, optional
            Controls wether interactive widgets should be shown with the plot
            to allow for temporal navigation. Only applicable if ``dataset``
            has multiple iterations.

        n: ``int``, optional
            Selects the index of the iteration to be shown initially. Only
            applicable if ``dataset`` has multiple iterations, .

        Returns
        ------
        :class:`nata.plots.Figure` or ``None``:
            Figure with plot built based on ``dataset``. Interactive widgets
            are shown with the figure if ``dataset`` has multiple iterations,
            in which case this method returns  ``None``.

        Examples
        --------
        To get a plot with default style properties in a new figure, simply
        call the ``.plot()`` method. The first two quantities in the dataset
        ``quantities`` dictionary will be represented in the horizontal and
        vertical plot axes, respectively. If a third quantity is available, it
        will be represented in colors.

        >>> from nata.containers import ParticleDataset
        >>> import numpy as np
        >>> arr = np.arange(30).reshape(1,10,3)
        >>> ds = ParticleDataset("path/to/file")
        >>> fig = ds.plot()

        The list of quantities in the dataset can be filtered with the
        :meth:`nata.containers.ParticleDataset.filter` method.

        >>> fig = ds.filter(quantities=["x1", "p1", "ene"]).plot()

    """

    p_plan = PlotPlan(
        dataset=dataset, style=filter_style(dataset.plot_type(), style),
    )

    a_plan = AxesPlan(
        axes=axes, plots=[p_plan], style=filter_style(Axes, style)
    )

    f_plan = FigurePlan(
        fig=fig, axes=[a_plan], style=filter_style(Figure, style)
    )

    if len(dataset) > 1:
        if inside_notebook():
            if interactive:
                f_plan.build_interactive(n)
            else:
                return f_plan[n].build()
        else:
            # TODO: remove last line from warn
            warn(
                f"Plotting only iteration with index n={str(n)}."
                + " Interactive plots of multiple iteration datasets are not"
                + " supported outside notebook environments."
            )
            return f_plan[n].build()

    else:
        return f_plan.build()



# -*- coding: utf-8 -*-
# flake8: noqa
"""
Copyright (C) 2017 Instituto Superior Tecnico

This file is part of the ZPIC Educational code suite

The ZPIC Educational code suite is free software: you can redistribute it
and/or modify it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the License,
or (at your option) any later version.

The ZPIC Educational code suite is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with the ZPIC Educational code suite. If not, see
<http://www.gnu.org/licenses/>.
"""

import sys

import numpy as np


class ZDF_Record:
    def __init__(self):
        self.pos = 0
        self.id = 0
        self.name = ""
        self.len = 0

    def version(self):
        return self.id & 0x0000FFFF

    def typeid(self):
        return self.id & 0xFFFF0000


class ZDF_Iteration:
    def __init__(self):
        self.name = ""
        self.n = 0
        self.t = 0.0
        self.tunits = ""


class ZDF_Grid_Axis:
    def __init__(self):
        self.name = ""
        self.type = 0
        self.min = 0.0
        self.max = 0.0
        self.label = ""
        self.units = ""


class ZDF_Grid_Info:
    def __init__(self):
        self.name = ""
        self.ndims = 0
        self.nx = []
        self.label = ""
        self.units = ""
        self.has_axis = 0
        self.axis = []


class ZDF_Part_Info:
    def __init__(self):
        self.name = ""
        self.label = ""
        self.nquants = 0
        self.quants = []
        self.qlabels = dict()
        self.qunits = dict()
        self.nparts = 0


class ZDF_Tracks_Info:
    def __init__(self):
        self.name = ""
        self.label = ""
        self.ntracks = 0
        self.ndump = 0
        self.niter = 0
        self.nquants = 0
        self.quants = []
        self.qlabels = []
        self.qunits = []


class ZDFfile:
    def __init__(self, file_name):
        self.__file = open(file_name, "rb")

        # Check magic number
        magic = self.__file.read(4)
        if magic != b"ZDF1":
            print("File is not a proper ZDF file, aborting", file=sys.stderr)
            self.__file.close

    def close(self):
        self.__file.close

    def __read_uint32(self):
        data = np.fromfile(self.__file, dtype="<u4", count=1)
        if data.size == 0:
            return False
        else:
            return data[0]

    def __read_int32(self):
        return np.fromfile(self.__file, dtype="<i4", count=1)[0]

    def __read_uint64(self):
        return np.fromfile(self.__file, dtype="<u8", count=1)[0]

    def __read_int64(self):
        return np.fromfile(self.__file, dtype="<i8", count=1)[0]

    def __read_float32(self):
        return np.fromfile(self.__file, dtype="<f4", count=1)[0]

    def __read_float64(self):
        return np.fromfile(self.__file, dtype="<f8", count=1)[0]

    def __read_string(self):

        length = self.__read_uint32()

        if length > 0:
            data = self.__file.read(length)
            fstring = data.decode()

            # Data is always written in blocks of 4 byt
            pad = ((length - 1) // 4 + 1) * 4 - length
            if pad > 0:
                self.__file.seek(pad, 1)
        else:
            fstring = ""

        return fstring

    def record_type(self, typeTag):
        typeID = typeTag & 0xFFFF0000

        types = {
            0x00010000: "int",
            0x00020000: "double",
            0x00030000: "string",
            0x00100000: "dataset",
            0x00110000: "cdset_start",
            0x00120000: "cdset_chunk",
            0x00130000: "cdset_end",
            0x00200000: "iteration",
            0x00210000: "grid_info",
            0x00220000: "part_info",
            0x00230000: "track_info",
        }

        if typeID in types:
            return types[typeID]
        else:
            return "unknown"

    # -----------------------------------------------------------------------------
    # Array datatypes
    # -----------------------------------------------------------------------------

    def __read_int32_arr(self, nx):
        size = np.prod(nx)
        data = np.fromfile(self.__file, dtype="<i4", count=size)
        data.shape = np.flip(nx)
        return data

    def __read_uint32_arr(self, nx):
        size = np.prod(nx)
        data = np.fromfile(self.__file, dtype="<u4", count=size)
        data.shape = np.flip(nx)
        return data

    def __read_int64_arr(self, nx):
        size = np.prod(nx)
        data = np.fromfile(self.__file, dtype="<i8", count=size)
        data.shape = np.flip(nx)
        return data

    def __read_uint64_arr(self, nx):
        size = np.prod(nx)
        data = np.fromfile(self.__file, dtype="<u8", count=size)
        data.shape = np.flip(nx)
        return data

    def __read_float32_arr(self, nx):
        size = np.prod(nx)
        data = np.fromfile(self.__file, dtype="<f4", count=size)
        data.shape = np.flip(nx)
        return data

    def __read_float64_arr(self, nx):
        size = np.prod(nx)
        data = np.fromfile(self.__file, dtype="<f8", count=size)
        data.shape = np.flip(nx)
        return data

    def __read_arr(self, dtype, nx):

        if dtype == 5:
            data = self.__read_int32_arr(nx)
        elif dtype == 3:
            data = self.__read_uint32_arr(nx)
        elif dtype == 7:
            data = self.__read_int64_arr(nx)
        elif dtype == 8:
            data = self.__read_uint64_arr(nx)
        elif dtype == 9:
            data = self.__read_float32_arr(nx)
        elif dtype == 10:
            data = self.__read_float64_arr(nx)
        else:
            print(
                "(*error*) ZDF: Data type not yet supported.", file=sys.stderr
            )
            data = False

        return data

    # -----------------------------------------------------------------------------
    # Low level interfaces
    # -----------------------------------------------------------------------------

    # -----------------------------------------------------------------------------
    # Read record header
    # -----------------------------------------------------------------------------

    def read_record(self, skip=False):

        rec = ZDF_Record()
        rec.pos = self.__file.tell()

        # Read record id and check for EOF
        rec.id = self.__read_uint32()

        if rec.id is False:
            # If end of file return false
            return False

        # Read name and length
        rec.name = self.__read_string()
        rec.len = self.__read_uint64()

        if skip:
            self.__file.seek(rec.len, 1)

        return rec

    def __record_skip(self, rec):
        self.__file.seek(rec.len, 1)

    # -----------------------------------------------------------------------------
    # Read string
    # -----------------------------------------------------------------------------

    def read_string(self, rec=False):
        if rec is False:
            rec = self.read_record()
        fstring = self.__read_string()
        return fstring

    # -----------------------------------------------------------------------------
    # Read iteration
    # -----------------------------------------------------------------------------

    def read_iteration(self, rec=False):
        if rec is False:
            rec = self.read_record()
        iteration = ZDF_Iteration()
        iteration.name = rec.name
        iteration.n = self.__read_int32()
        iteration.t = self.__read_float64()
        iteration.tunits = self.__read_string()
        return iteration

    # -----------------------------------------------------------------------------
    # Read grid info
    # -----------------------------------------------------------------------------

    def read_grid_info(self, rec=False):
        if rec is False:
            rec = self.read_record()

        # Maximum supported version
        max_version = 0x00000001

        # Get version
        version = rec.version()
        if version > max_version:
            print(
                "(*error*) ZDF: Grid info version is higher than supported.",
                file=sys.stderr,
            )
            print(
                "(*error*) ZDF: Please update the code to a newer version.",
                file=sys.stderr,
            )
            return False

        info = ZDF_Grid_Info()

        info.name = rec.name

        info.ndims = self.__read_uint32()
        info.nx = self.__read_uint64_arr(info.ndims)

        info.label = self.__read_string()
        info.units = self.__read_string()
        info.has_axis = self.__read_int32()

        if info.has_axis:
            for i in range(info.ndims):
                ax = ZDF_Grid_Axis()
                if version > 0:
                    ax.name = self.__read_string()
                else:
                    ax.name = "axis_{}".format(i)
                ax.type = self.__read_int32()
                ax.min = self.__read_float64()
                ax.max = self.__read_float64()
                ax.label = self.__read_string()
                ax.units = self.__read_string()
                info.axis.append(ax)

        return info

    # --------------------------------------------------------------------------
    # Read particle info
    # --------------------------------------------------------------------------

    def read_part_info(self, rec=False):
        if rec is False:
            rec = self.read_record()

        # Maximum supported version
        max_version = 0x00000002

        # Get version
        version = rec.version()
        if version > max_version:
            print(
                "(*error*) ZDF: Particles info version "
                + "is higher than supported.",
                file=sys.stderr,
            )
            print(
                "(*error*) ZDF: Please update the code to a newer version.",
                file=sys.stderr,
            )
            return False

        info = ZDF_Part_Info()
        info.name = rec.name
        info.label = self.__read_string()

        if version >= 1:
            # version 1
            info.nparts = self.__read_uint64()
            info.nquants = self.__read_uint32()

            for i in range(info.nquants):
                info.quants.append(self.__read_string())
            for q in info.quants:
                info.qlabels[q] = self.__read_string()
            for q in info.quants:
                info.qunits[q] = self.__read_string()

        else:
            # version 0
            info.nquants = self.__read_uint32()

            for i in range(info.nquants):
                info.quants.append(self.__read_string())

            # version 0 does not have label information
            for q in info.quants:
                info.qlabels[q] = q

            for q in info.quants:
                info.qunits[q] = self.__read_string()

            info.nparts = self.__read_uint64()

        return info

    # -----------------------------------------------------------------------------
    # Read track info
    # -----------------------------------------------------------------------------

    def read_track_info(self, rec=False):

        if rec is False:
            rec = self.read_record()

        # Maximum supported version
        max_version = 0x00000001

        # Get version
        version = rec.version()
        if version > max_version:
            print(
                "(*error*) ZDF: Tracks info version is higher than supported.",
                file=sys.stderr,
            )
            print(
                "(*error*) ZDF: Please update the code to a newer version.",
                file=sys.stderr,
            )
            return False

        info = ZDF_Tracks_Info()

        info.name = rec.name
        info.label = self.__read_string()
        info.ntracks = self.__read_uint32()
        info.ndump = self.__read_uint32()
        info.niter = self.__read_uint32()
        info.nquants = self.__read_uint32()

        for i in range(info.nquants):
            info.quants.append(self.__read_string())

        for i in range(info.nquants):
            info.qlabels.append(self.__read_string())

        for i in range(info.nquants):
            info.qunits.append(self.__read_string())

        # Iteration data is not supported so remove it from metadata
        info.nquants -= 1

        info.quants.pop(0)
        info.qlabels.pop(0)
        info.qunits.pop(0)

        return info

    # --------------------------------------------------------------------------
    # Read dataset
    # --------------------------------------------------------------------------

    def read_dataset(self, rec=False):

        if rec is False:
            rec = self.read_record()

        if self.record_type(rec.id) != "dataset":
            print(
                "(*error*) ZDF: Expected dataset record but found "
                + "{} instead.".format(self.record_type(rec.id)),
                file=sys.stderr,
            )
            return False

        # Maximum supported version
        max_version = 0x00000002

        # Get version
        version = rec.version()
        if version > max_version:
            print(
                "(*error*) ZDF: Dataset version is higher than supported.",
                file=sys.stderr,
            )
            print(
                "(*error*) ZDF: Please update the code to a newer version.",
                file=sys.stderr,
            )
            return False

        # Version 0x0001 includes id tag
        if version >= 1:
            id = self.__read_uint32()
        else:
            id = 0

        data_type = self.__read_int32()
        ndims = self.__read_uint32()
        nx = self.__read_uint64_arr(ndims)
        data = self.__read_arr(data_type, nx)

        return data

    # --------------------------------------------------------------------------
    # Read chunked dataset
    # --------------------------------------------------------------------------

    def read_cdset(self, rec=False, pos=0):

        if rec is False:
            rec = self.read_record()

        if self.record_type(rec.id) != "cdset_start":
            print(
                "(*error*) ZDF: Expected cdset_start record but found "
                + "{} instead.".format(self.record_type(rec.id)),
                file=sys.stderr,
            )
            return False

        # Maximum supported version
        max_version = 0x00000001

        # Get version
        version = rec.id & 0x0000FFFF
        if version > max_version:
            print(
                "(*error*) ZDF: Chunked dataset version is higher "
                + "than supported.",
                file=sys.stderr,
            )
            print(
                "(*error*) ZDF: Please update the code to a newer version.",
                file=sys.stderr,
            )
            return -1

        id = self.__read_uint32()
        data_type = self.__read_int32()
        ndims = self.__read_uint32()
        nx = self.__read_uint64_arr(ndims)

        # Create numpy array
        dt = {
            5: "int32",
            3: "uint32",
            7: "int64",
            8: "uint64",
            9: "float32",
            10: "float64",
        }

        data = np.zeros(np.flip(nx), dtype=dt[data_type])

        chunk_name = "{:#08}-chunk".format(id)
        end_name = "{:#08}-end".format(id)

        # Loop until the end record
        cdset_start_end = self.__file.tell()

        name = ""
        while name != end_name:

            # Read next record
            rec = self.read_record()

            # Check if end of file reached
            if rec is False:
                break

            name = rec.name

            if name == chunk_name:
                chunk_id = self.__read_uint32()
                count = self.__read_int64_arr(ndims)
                start = self.__read_int64_arr(ndims)
                stride = self.__read_int64_arr(ndims)

                chunk = self.__read_arr(data_type, count)

                if ndims == 1:
                    data[start[0] : start[0] + count[0] : stride[0]] = chunk
                elif ndims == 2:
                    data[
                        start[1] : start[1] + count[1] : stride[1],
                        start[0] : start[0] + count[0] : stride[0],
                    ] = chunk
                else:
                    data[
                        start[2] : start[2] + count[2] : stride[2],
                        start[1] : start[1] + count[1] : stride[1],
                        start[0] : start[0] + count[0] : stride[0],
                    ] = chunk

            else:

                self.__record_skip(rec)

        # If requested, position file pointer at the end of the cdset_start
        # record
        if pos == 1:
            self.__file.seek(cdset_start_end)

        return data

    # -----------------------------------------------------------------------------
    # Read arbitrary ZDF element
    # -----------------------------------------------------------------------------

    def read_element(self, rec=False, name=False, type_id=False):
        if rec is False:
            rec = self.read_record()

        if name:
            if name != rec.name:
                print(
                    "(*warning*) Requested name does not match record name",
                    file=sys.stderr,
                )
                print(
                    "(*warning*) expected '{}', found '{}".format(
                        name, rec.name
                    ),
                    file=sys.stderr,
                )
                self.__record_skip(rec)
                return False
        else:
            name = rec.name

        if type_id:
            if type_id != self.record_type(rec.id):
                print(
                    "(*warning*) Requested type does not match record type",
                    file=sys.stderr,
                )
                print(
                    "(*warning*) expected '{}', found '{}".format(
                        type_id, self.record_type(rec.id)
                    ),
                    file=sys.stderr,
                )
                self.__record_skip(rec)
                return False
        else:
            type_id = self.record_type(rec.id)

        if type_id == "int":
            data = self.__read_int32()
        elif type_id == "double":
            data = self.__read_float64()
        elif type_id == "string":
            data = self.__read_string()
        elif type_id == "dataset":
            data = self.read_dataset(rec=rec)
        elif type_id == "cdset_start":
            data = self.read_cdset(rec=rec)
        elif type_id == "cdset_chunk":
            print(
                "(*warning*) Dataset chunks are not meant to be read directly",
                file=sys.stderr,
            )
            self.__record_skip(rec)
            data = False
        elif type_id == "cdset_end":
            print("(*warning*) Dataset end marks have no data", file=sys.stderr)
            self.__record_skip(rec)
            data = False
        elif type_id == "iteration":
            data = self.read_iteration(rec=rec)
        elif type_id == "grid_info":
            data = self.read_grid_info(rec=rec)
        elif type_id == "part_info":
            data = self.read_part_info(rec=rec)
        elif type_id == "track_info":
            data = self.read_track_info(rec=rec)
        else:
            print("(*warning*) Unknown element type, skipping", file=sys.stderr)
            self.__record_skip(rec)
            data = False

        return data

    # -----------------------------------------------------------------------------
    # Read particle data
    # -----------------------------------------------------------------------------

    def read_part_data(self, quants):

        # Read all quantities
        data = dict()

        for q in quants:
            rec = self.read_record()

            # Sanity check, this should never happen
            if rec.name != q:
                print(
                    "(*error*) Expecting {} record, {} found".format(
                        q, rec.name
                    )
                )

            # Particle data can be either a standard dataset or a chunked
            # dataset
            type_id = self.record_type(rec.id)
            if type_id == "dataset":
                data[q] = self.read_dataset(rec=rec)
            elif type_id == "cdset_start":
                # Read chunked dataset and position file pointer after the
                # cdset_start record
                data[q] = self.read_cdset(rec=rec, pos=1)
            else:
                print(
                    "(*error*) Unable to read particle data, "
                    + "{} record found".format(type_id),
                    file=sys.stderr,
                )
                data[q] = None

        return data

    # -----------------------------------------------------------------------------
    # Read track data
    # -----------------------------------------------------------------------------

    def read_track_data(self, trackInfo):

        rec = self.read_record()
        if rec.name != "itermap":
            print(
                "(*error*) Expecting itermap record, {} found".format(rec.name),
                file=sys.stderr,
            )
        itermap = self.read_cdset(rec=rec, pos=1)

        rec = self.read_record()
        if rec.name != "data":
            print(
                "(*error*) Expecting data record, {} found".format(rec.name),
                file=sys.stderr,
            )
        data = self.read_cdset(rec=rec)

        trackNp = np.zeros(trackInfo.ntracks, dtype="<i8")
        for i in range(itermap.shape[0]):
            trackID = itermap[i, 0] - 1
            npoints = itermap[i, 1]
            trackNp[trackID] += npoints

        trackData = [None] * trackInfo.ntracks
        for i in range(trackInfo.ntracks):
            trackData[i] = np.zeros(
                [trackNp[i], trackInfo.nquants], dtype="<f4"
            )
            trackNp[i] = 0

        idx = 0
        for i in range(itermap.shape[0]):
            trackID = itermap[i, 0] - 1
            npoints = itermap[i, 1]
            (trackData[trackID])[
                trackNp[trackID] : trackNp[trackID] + npoints, :
            ] = data[idx : idx + npoints, :]
            trackNp[trackID] += npoints
            idx += npoints

        return trackData

    # -----------------------------------------------------------------------------
    # Retrieve list of file contents / print file contents
    # -----------------------------------------------------------------------------

    def list(self, printRec=False):
        # Position file after magic number
        self.__file.seek(4)

        rec_list = []
        while True:
            rec = self.read_record(skip=True)
            if rec is False:
                break
            else:
                rec_list.append(rec)

        if printRec and (len(rec_list) > 0):
            print("Position     Size(bytes)  Type         Name")
            print("-----------------------------------------------------")
            for rec in rec_list:
                print(
                    "{:#010x}   {:#010x}   {:11}  {}".format(
                        rec.pos, rec.len, self.record_type(rec.id), rec.name
                    )
                )

        return rec_list


# -----------------------------------------------------------------------------
# High level interfaces
# -----------------------------------------------------------------------------


class ZDF_Info:
    """ZDF File information"""

    def __init__(self):
        self.type = ""
        self.grid = None
        self.particles = None
        self.tracks = None
        self.iteration = None


def info(file_name):
    # Open file
    zdf = ZDFfile(file_name)

    # Check file type and read metadata
    info = ZDF_Info()
    info.type = zdf.read_string()
    if info.type == "grid":
        info.grid = zdf.read_grid_info()
        info.iteration = zdf.read_iteration()
    elif info.type == "particles":
        info.particles = zdf.read_part_info()
        info.iteration = zdf.read_iteration()
    elif info.type == "tracks-2":
        info.tracks = zdf.read_track_info()
    else:
        print(
            "File is not a valid ZDF grid, particles or tracks file",
            file=sys.stderr,
        )
        zdf.close()
        return False

    # Close file
    zdf.close()

    return info


def read(file_name):
    # Open file
    zdf = ZDFfile(file_name)

    # Check file type and read metadata
    info = ZDF_Info()

    info.type = zdf.read_string()
    if info.type == "grid":
        info.grid = zdf.read_grid_info()
        info.iteration = zdf.read_iteration()
        data = zdf.read_element()
    elif info.type == "particles":
        info.particles = zdf.read_part_info()
        info.iteration = zdf.read_iteration()
        data = zdf.read_part_data(info.particles.quants)
    elif info.type == "tracks-2":
        info.tracks = zdf.read_track_info()
        data = zdf.read_track_data(info.tracks)
    else:
        print(
            "File is not a valid ZDF grid, particles or tracks file",
            file=sys.stderr,
        )
        zdf.close()
        return False

    # Close file
    zdf.close()

    return (data, info)


def list(file_name):
    zdf = ZDFfile(file_name)
    zdf.list(True)
    zdf.close()



# -*- coding: utf-8 -*-
from typing import Union

import numpy as np

from nata.containers import GridDataset
from nata.plugins.register import register_container_plugin


@register_container_plugin(GridDataset, name="lineout")
def lineout_grid_dataset(
    dataset: GridDataset, fixed: Union[str, int], value: float,
) -> GridDataset:
    """Takes a lineout across a two-dimensional, single/multiple iteration\
       :class:`nata.containers.GridDataset`:

        Parameters
        ----------

        fixed: :class:``str`` or :class:``int``
            Selection of the axes along which the taken lineout is constant.

            * if it is a string, then it must match the ``name`` property of an
              existing grid axis in ``dataset``.
            * if it is an integer, then it must match the index of a grid axis
              in ``dataset`` (i.e. `0` or `1`).

        value: scalar
            Value between the minimum and maximum of the axes selected through
            ``fixed`` over which the lineout is taken.

        Returns
        ------
        :class:`nata.containers.GridDataset`:
            One-dimensional :class:`nata.containers.GridDataset`.

        Examples
        --------
        The following example shows how to obtain a lineout from a
        two-dimensional :class:`nata.containers.GridDataset`. Since no axes are
        attributed to the dataset in this example, they are automatically
        generated with no names, and ``fixed`` must be an integer.

        >>> from nata.containers import GridDataset
        >>> import numpy as np
        >>> arr = np.arange(25).reshape((5,5))
        >>> ds = GridDataset(arr[np.newaxis])
        >>> lo = ds.lineout(fixed=0, value=2)
        >>> lo.data
        array([10, 11, 12, 13, 14])

    """
    if len(dataset.grid_shape) != 2:
        raise ValueError(
            "Grid lineouts are only supported for two-dimensional grid datasets"
        )

    # get handle for grid axes
    axes = dataset.axes["grid_axes"]

    if isinstance(fixed, str):
        ax_idx = -1
        # get index based on
        for key, ax in enumerate(axes):
            if ax.name == fixed:
                ax_idx = key
                break
        if ax_idx < 0:
            raise ValueError(
                f"Axis `{fixed}` could not be found in dataset `{dataset}`"
            )
    else:
        ax_idx = fixed

    # build axis values
    axis = axes[ax_idx]

    if value < np.min(axis) or value > np.max(axis):
        raise ValueError(f"Out of range value for fixed `{fixed}`")

    values = np.array(axis)
    idx = (np.abs(values - value)).argmin()

    data = np.array(dataset)

    # get lineout
    if ax_idx == 0:
        lo_data = data[:, idx, :] if len(dataset) > 1 else data[idx, :]
        lo_axis = axes[1]

    elif ax_idx == 1:
        lo_data = data[:, :, idx] if len(dataset) > 1 else data[:, idx]
        lo_axis = axes[0]

    return GridDataset(
        lo_data if len(dataset) > 1 else lo_data[np.newaxis],
        name=f"{dataset.name}_lineout",
        label=dataset.label,
        unit=dataset.unit,
        grid_axes=[lo_axis],
        time=dataset.axes["time"],
        iteration=dataset.axes["iteration"],
    )



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import List
from typing import Optional

from nata.plots.data import PlotData

# from nata.plots.axes import Axes


@dataclass
class BasePlot:
    """Base class for plot types.

    Parameters
    ----------
    label: ``str``, optional
        Label of plot, used to identify the plot in the parent
        :class:`nata.plots.Axes` object legend. If not provided, it is inferred
        from the child dataset object.

    """

    # style properties
    label: Optional[str] = None

    # plot data object
    data: PlotData = None

    # parent axes object
    axes: Any = None

    # backend objects
    h: Any = field(init=False, repr=False, default=None)

    def __post_init__(self):
        # set defaults on parent axes object
        axes = self.axes

        if axes.xlim is None:
            axes.xlim_auto = True
            axes.xlim = self._default_xlim()

        if axes.ylim is None:
            axes.ylim_auto = True
            axes.ylim = self._default_ylim()

        if axes.xlabel is None:
            axes.xlabel_auto = True
            axes.xlabel = self._default_xlabel()

        if axes.ylabel is None:
            axes.ylabel_auto = True
            axes.ylabel = self._default_ylabel()

        if axes.title is None:
            axes.title_auto = True
            axes.title = self._default_title()

        # set default label
        if self.label is None:
            self.label = self._default_label()

        self.build_canvas()

    def _default_xlim(self):
        return ()

    def _default_ylim(self):
        return ()

    def _default_xlabel(self, units=True):
        return ""

    def _default_ylabel(self, units=True):
        return ""

    def _default_title(self):
        return ""

    def _default_label(self):
        return ""

    def _xunits(self):
        return ""

    def _yunits(self):
        return ""

    def build_canvas(self):
        pass

    def clear(self):
        pass

    @classmethod
    def style_attrs(self) -> List[str]:
        return [
            "label",
        ]



# -*- coding: utf-8 -*-
from typing import Optional

import numpy as np

from nata.containers import DatasetCollection
from nata.plots.axes import Axes
from nata.plots.figure import Figure
from nata.plots.helpers import filter_style
from nata.plots.plans import AxesPlan
from nata.plots.plans import FigurePlan
from nata.plots.plans import PlotPlan
from nata.plugins.register import register_container_plugin
from nata.utils.env import inside_notebook


@register_container_plugin(DatasetCollection, name="plot")
def plot_collection(
    collection: DatasetCollection,
    order: Optional[list] = list(),
    styles: Optional[dict] = dict(),
    interactive: bool = True,
    n: int = 0,
) -> Figure:

    # check if collection is not empty
    if not collection.store:
        raise ValueError("Collection is empty.")

    # check if time and iteration arrays are equal
    for check in ["iteration", "time"]:
        arr = [
            np.array(dataset.axes[check])
            for dataset in collection.store.values()
        ]
        if not all(np.array_equal(arr[0], i) for i in arr):
            raise ValueError(
                f"Attribute `{check}` is not the same for all datasets "
                + "in the collection."
            )

    f_a = []

    for dataset in collection.store.values():
        i_style = (
            styles[dataset.name] if dataset.name in styles.keys() else None
        )

        p_plan = PlotPlan(
            dataset=dataset, style=filter_style(dataset.plot_type(), i_style),
        )

        a_plan = AxesPlan(
            axes=None, plots=[p_plan], style=filter_style(Axes, i_style)
        )

        f_a.append(a_plan)

    f_plan = FigurePlan(
        fig=None,
        axes=f_a,
        style=filter_style(
            Figure, styles["fig"] if "fig" in styles.keys() else None
        ),
    )

    if len(dataset) > 1 and inside_notebook() and interactive:
        f_plan.build_interactive(n)

    else:
        return f_plan.build()



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import List
from typing import Optional

import numpy as np

from nata.plots.types import BasePlot


@dataclass
class LinePlot(BasePlot):
    """Line plot class.

    Parameters
    ----------
    ls: ``str``, optional
        Linestyle of the line. See
        :meth:`matplotlib.lines.Line2D.set_linestyle` for available options.

    lw: ``float``, optional
        Line width in points. If not provided, defaults to ``1``.

    color: ``str``, optional
        Color of the line. See :mod:`matplotlib.colors` for available options.

    alpha: ``float``, optional
        Line alpha value. Must be between ``0`` and ``1``.

    marker: ``str``, optional
        Marker to be used in defined line points. See :mod:`matplotlib.markers`
        for available options.

    ms: ``float``, optional
        Marker size in points.

    antialiased: ``bool``, optional
        Controls wether the plot should be antialiased. If not provided,
        defaults to ``True``.

    """

    ls: Optional[str] = None
    lw: Optional[float] = 1
    color: Optional[str] = None
    alpha: Optional[float] = None
    marker: Optional[str] = None
    ms: Optional[float] = None
    antialiased: Optional[bool] = True

    def _default_xlim(self):
        return (self.data.axes[0].min, self.data.axes[0].max)

    def _default_ylim(self):
        return (np.min(self.data.data), np.max(self.data.data))

    def _default_xlabel(self, units=True):
        return self.data.axes[0].get_label(units)

    def _default_ylabel(self, units=True):
        return self.data.get_label(units)

    def _default_title(self):
        return self.data.get_time_label()

    def _default_label(self):
        return self.data.get_label(units=False)

    def _xunits(self):
        return f"${self.data.axes[0].units}$" if self.data.axes[0].units else ""

    def _yunits(self):
        return f"${self.data.units}$" if self.data.units else ""

    def build_canvas(self):
        # get plot axes and data
        x = self.data.axes[0].data
        y = self.data.data

        # build plot
        self.h = self.axes.ax.plot(
            x,
            y,
            ls=self.ls,
            lw=self.lw,
            c=self.color,
            alpha=self.alpha,
            marker=self.marker,
            ms=self.ms,
            label=self.label,
            antialiased=self.antialiased,
        )

    @classmethod
    def style_attrs(cls) -> List[str]:
        return [
            "ls",
            "lw",
            "color",
            "alpha",
            "marker",
            "ms",
            "antialiased",
        ] + BasePlot.style_attrs()



# -*- coding: utf-8 -*-
"""Nata types

This file contain all the different types available with nata. It is meant to
be used for typechecking and type annotation. In addition, type checking at
runtime is supported.
"""
import sys
from pathlib import Path
from typing import AbstractSet
from typing import Any
from typing import Dict
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np

# "Protocol" and "runtime_checkable" are builtin for 3.8+
# otherwise use "typing_extension" package
if sys.version_info >= (3, 8):
    from typing import Protocol
    from typing import runtime_checkable
    from typing import TypedDict
else:
    from typing_extensions import Protocol
    from typing_extensions import runtime_checkable
    from typing_extensions import TypedDict


def is_basic_indexing(key: Any):
    indexing = np.index_exp[key]
    passes = []
    for ind in indexing:
        if isinstance(ind, (int, slice)):
            passes.append(True)
        elif ind is Ellipsis:
            passes.append(True)
        elif ind is np.newaxis:
            passes.append(True)
        else:
            passes.append(False)

    if all(passes):
        return True
    return False


#: Scalars and numbers
Number = Union[float, int]

#: Type which can be supplied to `numpy.array` and the resulting output is an
#: array
ArrayLike = Union[np.ndarray, Sequence[Number]]

#: Type for basic indexing
BasicIndex = Union[int, slice]

#: Type for basic indexing
BasicIndexing = Union[BasicIndex, Tuple[BasicIndex, ...]]

#: Type for file location
FileLocation = Union[Path, str]


@runtime_checkable
class BackendType(Protocol):
    """General type for retreiving data.

    The :class:`.BackendType` characterizes a general behavior of a backend.
    It is a Protocol which characterizes available attributes and their
    corresponding type annotation.
    """

    #: Name of the backend. The name can be chosen individually and is used
    #: for providing users with information about the underlaying data
    #: storage. It should follow the convention
    #: ``"CODENAME_VERSION_DATATYPE_STORAGE"``, e.g.
    #: ``"osiris_4.4.4_grid_hdf5"``
    name: str

    #: Location of the data. This attribute can be used inside a backend to
    #: point to data, either to open a file or to retrieve it.
    location: Optional[Union[str, Path]]

    @staticmethod
    def is_valid_backend(location: Union[Path, str]) -> bool:
        """Determine if a backend is a valid backend.

        Parameters
        ----------
        location : :obj:`str` or :obj:`pathlib.Path`
            Checks if a location can be passed to backend to initiate it.

        Returns
        -------
        out : :obj:`bool`
            Returns ``True`` if ``location`` is valid input parameter for
            instantiation and ``False`` otherwise.
        """
        ...


@runtime_checkable
class GridBackendType(BackendType, Protocol):
    """Backend representing a grid.

    `GridBackendType` is a protocol with the purpose of characterizing
    attributes being available for object to be recognized as a
    `GridBackendType`. Reading data is not part of this protocol but is
    characterized by `GridDataReader` which extends this protocol.
    """

    #: Name of the dataset. It has to be identifiable, e.g.
    #: ``"some_dataset_name"``.
    dataset_name: str

    #: Descriptive label of the dataset. Can be an arbitrary string, e.g.
    #: ``"some long label with space"``.
    dataset_label: str

    #: Unit of the correspinding grid. Can be an string including some latex
    #: symbols, e.g. ``"m_e c \\omega_p e^{-1}"``.
    dataset_unit: str

    #: A sequence of strings for each grid axis. Each string of the sequence
    #: has to be identifiable, e.g. ``["axis0", "axis1", "axis2"]``.
    axes_names: Sequence[str]
    #: A sequence of strings for each grid axis. Each string of the sequence
    #: is a descriptive label for each axis, e.g.
    #: ``["some axis 0", "some axis 1", "some axis 2"]``.
    axes_labels: Sequence[str]
    #: A sequence of strings for each grid axis. Each string of the sequence
    #: is the unit for each axis including some latex symbols, e.g.
    #: ``["c / \\omega_p", "mm", "\\omega_p^{-1}"]``.
    axes_units: Sequence[str]
    #: An array representing the lower limits of each grid axis.
    axes_min: np.ndarray
    #: An array representing the upper limits of each grid axis.
    axes_max: np.ndarray

    #: Associated iteration step of the underlaying data.
    iteration: int
    #: ::Associated time step of the underlaying data in code units.
    time_step: float
    #: Unit for time. Can be an arbitrary string, e.g. ``"1 / \\omega_p"``.
    time_unit: str

    #: Tuple of grid array dimensions. Corresponds to
    #: `numpy.ndarray.shape`.
    shape: Tuple[int, ...]
    #: Data type object of the grid array. Corresponds to `numpy.dtype`.
    dtype: np.dtype
    #: Dimensionality of the grid. Corresponds to `numpy.ndarray.ndim`
    ndim: int


@runtime_checkable
class GridDataReader(GridBackendType, Protocol):
    """Extended backend which handles grid data reading"""

    def get_data(self, indexing: Optional[BasicIndexing] = None) -> np.ndarray:
        """Routine for reading underlaying grid data.

        Parameters
        ----------
        indexing : `int`, `slice`, \
                   :class:`typing.Tuple[Union[slice, int], ...]]`, optional
            Optional indexing for reading a section of the grid. Any `basic
            slicing and indexing  <https://numpy.org/doc/stable/reference/\
            arrays.indexing.html#basic-slicing-and-indexing>`_ can be passed
            here.

        Returns
        -------
        out : :class:`numpy.ndarray`
            Data array of the underlaying grid.
        """
        ...


@runtime_checkable
class ParticleBackendType(BackendType, Protocol):
    """Backend representing a particles.

    `ParticleBackendType` is a protocol with the purpose of characterizing
    attributes being available for object to be recognized as a
    `ParticleBackendType`. Reading data is not part of this protocol but is
    characterized by `ParticleDataReader` which extends this protocol.
    """

    #: Name of the dataset. It has to be identifiable, e.g.
    #: ``"some_dataset_name"``.
    dataset_name: str
    #: Number of particles for a given backend.
    num_particles: int

    #: A sequence of strings for each quantity stored in a backend. Each
    #: string of the sequence has to be identifiable, e.g. ``["quant0",
    #: "quant1", "quant2"]``.
    quantity_names: Sequence[str]
    #: A sequence of strings for each quantity stored in a backend. Each
    #: string of the sequence is a descriptive label for each quantity, e.g.
    #: ``["some quantity 0", "some quantity 1", "some quantity 2"]``.
    quantity_labels: Sequence[str]
    #: A sequence of strings for the unit of each quantity. Each string of
    #: the sequence is the unit for each quantity including some latex symbols,
    #: e.g. ``["m_e", "c / \\omega_p"]``.
    quantity_units: Sequence[str]

    #: Associated iteration step of the underlaying data.
    iteration: int
    #: Associated time step of the underlaying data in code units.
    time_step: float
    #: Unit for time. Can be an arbitrary string, e.g. ``"1 / \\omega_p"``.
    time_unit: str

    #: Structured type of the underlaying particle backend. Field names
    #: correspond to quantity names, and the type corresponds to the array type.
    dtype: np.dtype


@runtime_checkable
class ParticleDataReader(ParticleBackendType, Protocol):
    """Extended backend which handles particle data reading"""

    def get_data(
        self,
        indexing: Optional[BasicIndex] = None,
        fields: Optional[Union[str, Sequence[str]]] = None,
    ) -> np.ndarray:
        """Routine for reading underlaying grid data.

        Parameters
        ----------
        indexing : `int`, `slice`, \
                   :class:`typing.Tuple[Union[slice, int], ...]]`, optional
            Optional indexing for reading a section of the grid. Any `basic
            slicing and indexing  <https://numpy.org/doc/stable/reference/\
            arrays.indexing.html#basic-slicing-and-indexing>`_ can be passed
            here.
        fields: `str`, :class:`typing.Sequence[str]`, optional
            Optional field or Sequence of fields to read out. Each field
            correspond to a quantity stored in the backend.

        Returns
        -------
        out : :class:`numpy.ndarray`
            Data array of the underlaying particle array.
        """
        ...


@runtime_checkable
class DatasetType(Protocol):
    """Base protocol for datasets.

    Each dataset has one private attribute which stores the information about
    available backends and the possibility of storing/removing backends to
    the backend store.
    """

    #: Storage of available backends for a dataset.
    _backends: AbstractSet[BackendType]

    #: Name of associated backend.
    backend: Optional[str]

    @classmethod
    def add_backend(cls, backend: BackendType) -> None:
        """Attach a new backend to backend store.

        Parameters
        ----------
        backend : `BackendType`
            Backend which will be stored in `_backends`.
        """
        ...

    @classmethod
    def remove_backend(cls, backend: BackendType) -> None:
        """Remove an attached backend from backend store.

        Parameters
        ----------
        backend : `BackendType`
            Backend which will be removed from `._backends`.
        """
        ...

    @classmethod
    def is_valid_backend(cls, backend: BackendType) -> bool:
        """Checks if a backend is a valid backend for a dataset.

        Parameters
        ----------
        backend : `BackendType`
            Backend which will be checked if it is a valid backend for a
            dataset.

        Returns
        -------
        out : `bool`
            `True` if a backend is a valid backend for a Dataset and `False`
            otherwise.
        """
        ...

    @classmethod
    def get_backends(cls) -> Dict[str, BackendType]:
        """Obtain information over stored backends.

        Returns a `dict` with information of the stored backends inside a
        dataset. The keys are of type `str` and are the names of the
        backends. The values of the dictionary are the backends
        `BackendType`.
        """
        ...

    def append(self, other: "DatasetType") -> None:
        """Appends another dataset.

        Parameters
        ----------
        other : `DatasetType`
            The other dataset which will be appended.
        """
        ...

    def equivalent(self, other: "DatasetType") -> bool:
        """Checks for equivalence of two datasets.

        Parameters
        ----------
        other : `DatasetType`
            The other dataset which will be checked.

        Returns
        -------
        out : `bool`
            Returns `True` if an instance of a datasets is equivalent with
            another datasets, `False` otherwise.
        """
        ...


@runtime_checkable
class HasArrayInterface(Protocol):
    """Base protocol for an object to be 'characterized' as an array."""

    #: Represents the stored data for an object. It is of type `numpy.ndarray`.
    data: np.ndarray

    #: Data type object of the stored data. It is of type `numpy.dtype`.
    dtype: np.dtype
    #: Shape of the underlaying stored data. Similar to `numpy.ndarray.shape`
    shape: Tuple[int, ...]
    #: Dimensionality of the underlaying stored data. Similar to
    #: `numpy.ndarray.ndim`
    ndim: int

    def __array__(self, dtype: Optional[np.dtype] = None) -> np.ndarray:
        """Array interface of an object.

        Method which will be called when `numpy.array` or similar function
        called with the object provided as an input.
        """
        ...


@runtime_checkable
class AxisType(Protocol):
    """Base protocol for an axis."""

    #: Name of the axis. It has to be identifiable, e.g.
    #: ``"some_axis_name"``.
    name: str
    #: Descriptive label of the axis. Can be an arbitrary string, e.g.
    #: ``"some long label for an axis"``.
    label: str
    #: Unit of the correspinding axis. Can be an string including some latex
    #: symbols, e.g. ``"m_e c \\omega_p e^{-1}"``.
    unit: str
    #: Dimensionality of an axis.
    axis_dim: int

    def append(self, other: "AxisType") -> None:
        """Appends another axis.

        Parameters
        ----------
        other : `AxisType`
            The other axis which will be appended.
        """
        ...

    def equivalent(self, other: "AxisType") -> bool:
        """Checks for equivalence of two axes.

        Parameters
        ----------
        other : `AxisType`
            The other axis which will be checked.

        Returns
        -------
        out : `bool`
            Returns `True` if an instance of a  is equivalent
            with another particle quantity, `False` otherwise.
        """
        ...


@runtime_checkable
class GridAxisType(AxisType, Protocol):
    #: Axis type of a grid axis.
    axis_type: str


@runtime_checkable
class QuantityType(Protocol):
    """Base protocol for particle quantities."""

    #: Name of the quantity. It has to be identifiable, e.g.
    #: ``"some_quantity_name"``.
    name: str
    #: Descriptive label of the particle quantity. Can be an arbitrary string,
    #: e.g. ``"some long label for a particle quantity"``.
    label: str
    #: Unit of the correspinding particle quantity. Can be an string including
    #: some latex symbols, e.g. ``"m_e c \\omega_p e^{-1}"``.
    unit: str

    def append(self, other: "QuantityType") -> None:
        """Appends another a particle quantity.

        Parameters
        ----------
        other : `QuantityType`
            The other quantity which will be appended.
        """
        ...

    def equivalent(self, other: "QuantityType") -> bool:
        """Checks for equivalence of two particle quantaties.

        Parameters
        ----------
        other : `QuantityType`
            The other particle quantity which will be checked.

        Returns
        -------
        out : `bool`
            Returns `True` if an instance of a particle quantity is equivalent
            with another particle quantity, `False` otherwise.
        """
        ...


class GridDatasetAxes(TypedDict):
    """Typed dictionary containing axes for grid datasets.

    Typed dictionary `typing.TypedDict` which correspond to a `dict` at
    runtime are merely there for type checking. The attributes correspond to
    required keys inside a dictionary.
    """

    #: Axis to store iteration information.
    iteration: Optional[AxisType]
    #: Axis to store time information.
    time: Optional[AxisType]
    #: Sequence of `GridAxisType` representing an axis for each grid dimension.
    grid_axes: Sequence[GridAxisType]


@runtime_checkable
class GridDatasetType(DatasetType, Protocol):
    """Base protocol for GridDatasets.

    Extends `DatasetType` to include additional information for grids.
    """

    #: Name of the grid dataset. It has to be identifiable, e.g.
    #: ``"some_grid_name"``.
    name: str
    #: Descriptive label of the grid. Can be an arbitrary string, e.g.
    #: ``"some long label for a grid"``.
    label: str
    #: Unit of the correspinding grid. Can be an string including some latex
    #: symbols, e.g. ``"m_e c \\omega_p e^{-1}"``.
    unit: str

    #: Axes for `GridDatasetType`. It is a dictionary of type `GridDatasetAxes`.
    axes: GridDatasetAxes
    #: Shape of the grid. The grid shape corresponds to the underlaying grid
    #: and does not include temporal infortions.
    grid_shape: Tuple[int, ...]


class ParticleDatasetAxes(TypedDict):
    """Typed dictionary containing axes for particle datasets.

    Typed dictionary `typing.TypedDict` which correspond to a `dict` at
    runtime are merely there for type checking. The attributes correspond to
    required keys inside a dictionary.
    """

    #: Axis to store iteration information.
    iteration: Optional[AxisType]
    #: Axis to store time information.
    time: Optional[AxisType]


@runtime_checkable
class ParticleDatasetType(DatasetType, Protocol):
    """Base protocol for ParticleDatasets.

    Extends `DatasetType` to include additional information for particles.
    """

    #: Name of the particle dataset. It has to be identifiable, e.g.
    #: ``"some_particle_species_name"``.
    name: str

    #: Mapping storing information about stored quantaties. Keys represent are
    #: names for quantities and values are Store
    quantities: Mapping[str, QuantityType]
    #: Axes for `ParticleDatasetType`. It is a dictionary of type
    #: `ParticleDatasetAxes`.
    axes: ParticleDatasetAxes



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import List
from typing import Optional

import numpy as np


@dataclass
class PlotDataAxis:
    name: str = ""
    label: Optional[str] = None
    units: Optional[str] = None
    data: Optional[np.ndarray] = None

    @property
    def min(self) -> float:
        return np.min(self.data)

    @property
    def max(self) -> float:
        return np.max(self.data)

    def get_label(self, units=True) -> str:
        label = ""
        if self.label:
            label += f"${self.label}$"
            if units and self.units:
                label += f" $\\left[{self.units}\\right]$"
        return label


@dataclass
class PlotData:
    data: np.ndarray
    axes: List[PlotDataAxis]
    name: str = ""
    label: Optional[str] = None
    units: Optional[str] = None
    time: Optional[float] = None
    time_units: Optional[str] = None

    def get_label(self, units=True) -> str:
        label = ""
        if self.label:
            label += f"${self.label}$"
            if units and self.units:
                label += f" $\\left[{self.units}\\right]$"
        return label

    def get_time_label(self) -> str:
        label = ""
        if self.time is not None:
            label += f"Time = ${self.time:.2f}$"
            if self.time_units:
                label += f" $\\left[{self.time_units}\\right]$"

        return label



# -*- coding: utf-8 -*-
from typing import MutableSequence
from typing import Optional
from typing import Sequence


def sort_particle_quantities(
    names: MutableSequence[str], order: Optional[Sequence[str]] = None
):
    if order is None:
        return sorted(names)
    else:
        sorted_ = []
        for key in order:
            filtered_names = filter(lambda elem: elem.startswith(key), names)
            for s in sorted(filtered_names):
                sorted_.append(s)
                names.remove(s)

        return sorted_ + sorted(names)



# -*- coding: utf-8 -*-
import re

from numpy import ndarray


def array_format(data: ndarray):
    if data.ndim == 0:
        return str(data)
    elif len(data) < 4:
        return "[" + ", ".join(str(i) for i in data) + "]"
    else:
        return (
            "["
            + ", ".join(str(i) for i in data[:2])
            + ", ... "
            + str(data[-1])
            + "]"
        )


def make_as_identifier(s: str):
    # Remove leading characters until we find a letter or underscore
    s = re.sub("^[^a-zA-Z]+", "", s)

    # replace spaces by underscore
    s = s.replace(" ", "_")

    # Remove invalid characters
    s = re.sub("[^0-9a-zA-Z_]", "", s)

    return s


def make_identifiable(s: str) -> str:
    # Remove leading characters until we find a letter or underscore
    s = re.sub("^[^a-zA-Z]+", "", s)

    # replace spaces by underscore
    s = s.replace(" ", "_")

    # Remove invalid characters
    s = re.sub("[^0-9a-zA-Z_]", "", s)

    return s



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from typing import List
from typing import Optional

import matplotlib.colors as clr
import numpy as np

from nata.plots.types import BasePlot


@dataclass
class ScatterPlot(BasePlot):
    """Color plot class.

    Parameters
    ----------
    s: ``float``, optional
        Marker size in in points**2. If not provided, defaults to ``0.1``

    c: ``str``, optional
        Color of the markers. See :mod:`matplotlib.colors` for available
        options.

    marker: ``str``, optional
        Marker style. See :mod:`matplotlib.markers` for available options.

    alpha: ``float``, optional
        Marker alpha value. Must be between ``0`` and ``1``.

    vmin: ``float``, optional
        Minimum of the colorbar axis. If not provided, it is
        inferred from the dataset represented in the plot.

    vmax: ``float``, optional
        Same as ``vmin`` for the maximum of the colorbar axis.

    cb_title: ``str``, optional
        Colorbar title. If not provided, it is inferred from the dataset
        represented in the plot.

    cb_scale: ``{'linear','log', 'symlog'}``, optional
        Scale of the colorbar. If not provided, defaults to ``'linear'``.

    cb_map: ``str``, optional
        Colormap used to represent the data. See
        :func:`matplotlib.pyplot.colormaps` for available options. If not
        provided, defaults to ``rainbow``.

    cb_linthresh: ``float``, optional
        Range within which the colorbar axis is linear. Applicable only when
        ``cb_scale`` is set to ``'symlog'``. If not provided, defaults to
        ``1e-5``.

    Notes
    -----
    All colorbar parameters are only applicable if the dataset represented in
    the plot has a quantity to be represented in color. In this case, ``c`` is
    overriden if set.

    """

    s: Optional[float] = 0.1
    c: Optional[str] = None
    marker: Optional[str] = None
    alpha: Optional[float] = None
    vmin: Optional[float] = None
    vmax: Optional[float] = None
    cb_map: Optional[str] = "rainbow"
    cb_scale: Optional[str] = "linear"
    cb_linthresh: Optional[float] = 1e-5
    cb_title: Optional[str] = None

    def __post_init__(self):
        if self.cb_title is None and len(self.data.axes) > 2:
            self.cb_title = self.data.axes[2].get_label(units=True)

        if self.has_cb:
            self.c = self.data.data[2]

        super().__post_init__()

    @property
    def has_cb(self):
        return len(self.data.axes) > 2

    def _default_xlim(self):
        return (np.min(self.data.data[0]), np.max(self.data.data[0]))

    def _default_ylim(self):
        return (np.min(self.data.data[1]), np.max(self.data.data[1]))

    def _default_xlabel(self, units=True):
        return self.data.axes[0].get_label(units)

    def _default_ylabel(self, units=True):
        return self.data.axes[1].get_label(units)

    def _default_title(self):
        return self.data.get_time_label()

    def _default_label(self):
        return self.data.get_label(units=False)

    def _xunits(self):
        return f"${self.data.axes[0].units}$" if self.data.axes[0].units else ""

    def _yunits(self):
        return f"${self.data.axes[1].units}$" if self.data.axes[1].units else ""

    def build_canvas(self):
        # get plot axes and data
        x = self.data.data[0]
        y = self.data.data[1]

        # build color map norm
        if self.cb_scale == "log":
            self.cb_norm = clr.LogNorm(vmin=self.vmin, vmax=self.vmax)
        elif self.cb_scale == "symlog":
            self.cb_norm = clr.SymLogNorm(
                vmin=self.vmin,
                vmax=self.vmax,
                linthresh=self.cb_linthresh,
                base=10,
            )
        else:
            self.cb_norm = clr.Normalize(vmin=self.vmin, vmax=self.vmax)

        # build plot
        self.h = self.axes.ax.scatter(
            x,
            y,
            s=self.s,
            c=self.c,
            marker=self.marker,
            alpha=self.alpha,
            label=self.label,
            cmap=self.cb_map,
            norm=self.cb_norm
            # antialiased=self.antialiased,
        )

        if self.has_cb:
            self.axes.init_colorbar(plot=self)

    @classmethod
    def style_attrs(cls) -> List[str]:
        return [
            "s",
            "c",
            "marker",
            "alpha",
            "vmin",
            "vmax",
            "cb_map",
            "cb_scale",
            "cb_linthresh",
            "cb_title",
        ] + BasePlot.style_attrs()



# -*- coding: utf-8 -*-
from copy import copy
from dataclasses import dataclass
from dataclasses import field
from math import ceil
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
from pkg_resources import resource_filename

from nata.plots.axes import Axes


@dataclass
class Figure:
    """Container of parameters and child objects (including plotting\
    backend-related objects) relevant to draw a figure.

    Parameters
    ----------
    figsize: ``tuple`` of ``float``, optional
        Tuple containing the width and height of the figure canvas in
        inches. If not provided, defaults to ``(6,4)``.

    nrows: ``int``, optional
        Number of rows available for figure axes. If not provided, defaults
        to ``1``.

    ncols: ``int``, optional
        Number of columns available for figure axes. If not provided,
        defaults to ``1``.

    style: ``{'light', 'dark'}``, optional
        Selection of standard nata style. If not provided, defaults to
        ``'light'``.

    fname: ``str``, optional
        Path to file with custom plotting backend parameters.

    rc: ``dict``, optional
        Dictionary with custom plotting backend parameters. Overrides
        parameters given in ``fname``.

    """

    figsize: Optional[Tuple[float]] = None
    nrows: Optional[int] = 1
    ncols: Optional[int] = 1
    style: Optional[str] = "light"
    fname: Optional[str] = None
    rc: Optional[Dict[str, Any]] = None

    # backend objects
    fig: Any = field(init=False, repr=False, default=None)

    # child axes objects
    _axes: List[Axes] = field(init=False, repr=False, default_factory=list)

    def __post_init__(self):
        # set plotting style
        self.set_style()

        # open figure object
        self.open()

    def set_style(self):
        if not self.fname:
            self.fname = resource_filename(
                __name__, "styles/" + self.style + ".rc"
            )

    # TODO: generalize methods for arbitrary backend
    def open(self):
        with mpl.rc_context(fname=self.fname, rc=self.rc):
            self.fig = plt.figure(figsize=self.figsize)

            if self.figsize is None:
                size = self.fig.get_size_inches()
                self.fig.set_size_inches(
                    size[0] * self.ncols, size[1] * self.nrows
                )

    def close(self):
        plt.close(self.fig)

    def reset(self):
        self.close()
        self.open()

    def show(self):
        """Shows the figure."""

        with mpl.rc_context(fname=self.fname, rc=self.rc):
            dummy = plt.figure()
            new_manager = dummy.canvas.manager
            new_manager.canvas.figure = self.fig

            self.fig.tight_layout()
            plt.show()

    def _repr_html_(self):
        """Calls :meth:`nata.plots.Figure.show`."""
        self.show()

    # TODO: generalize this for arbitrary backend
    def save(
        self, path, format: Optional[str] = None, dpi: Optional[float] = 150
    ):
        """Saves the figure to a file.

        Parameters
        ----------
            path: ``tuple`` of ``float``, optional
                Path in which to store the file.

            format: ``str``, optional
                File format, e.g. ``'png'``, ``'pdf'``, ``'svg'``. If not
                provided, the output format is inferred from the extension of
                ``path``.

            dpi: ``float``, optional
                Resolution in dots per inch. If not provided, defaults to
                ``150``.

        """
        with mpl.rc_context(fname=self.fname, rc=self.rc):
            self.fig.tight_layout()
            self.fig.savefig(path, dpi=dpi, bbox_inches="tight")

    def copy(self):

        self.close()

        new = copy(self)
        new.open()

        for axes in new._axes:
            axes.fig = new

        return new

    def add_axes(self, style=dict()):

        new_index = len(self._axes) + 1

        if new_index > (self.nrows * self.ncols):
            # increase number of rows
            # TODO: really?
            self.nrows += 1

            if self.figsize is None:
                size = self.fig.get_size_inches()
                self.fig.set_size_inches(
                    size[0], size[1] * self.nrows / (self.nrows - 1)
                )

            for axes in self._axes:
                axes.redo_plots()

        axes = Axes(fig=self, index=new_index, **style)
        self._axes.append(axes)

        return axes

    def __mul__(self, other):
        """Combines two figures into one by superimposing the plots in axes with
        matching indices.
        """

        new = copy(self)

        for key, axes in new.axes.items():

            if key in other.axes:
                for plot in other.axes[key].plots:
                    axes.add_plot(plot=plot)

            axes.redo_plots()

        new.close()

        return new

    def __add__(self, other):
        """Combines two figures into one by adding new axes."""

        new = self.copy()

        new.nrows = ceil((len(new._axes) + len(other._axes)) / new.ncols)

        if new.figsize is None:
            size = new.fig.get_size_inches()
            new.fig.set_size_inches(
                size[0], size[1] * new.nrows / (new.nrows - 1)
            )

        for axes in new._axes:
            axes.redo_plots()

        for axes in other._axes:
            # get a copy of old axes
            new_axes = axes.copy()

            # reset parent figure object
            new_axes.fig = new

            # redo plots in new axes
            new_axes.index = len(new._axes) + 1
            new_axes.redo_plots()

            # add axes to new list
            new._axes.append(new_axes)

        new.close()

        return new

    @property
    def axes(self) -> dict:
        """Dictionary of child `nata.plots.Axes` objects, where the key
        to each axes is its ``index`` property
        """
        return {axes.index: axes for axes in self._axes}

    @classmethod
    def style_attrs(self) -> List[str]:
        return [
            "figsize",
            "nrows",
            "ncols",
            "style",
            "fname",
            "rc",
        ]



# -*- coding: utf-8 -*-
import pathlib
from typing import Union

from nata.containers import DatasetCollection


class _FileList:
    def __init__(
        self, entrypoint: Union[str, pathlib.Path], recursive: bool = True
    ):
        self._entrypoint = (
            pathlib.Path(entrypoint)
            if not isinstance(entrypoint, pathlib.Path)
            else entrypoint
        )
        self._recursive = recursive

        if not self._entrypoint.exists():
            if not self._entrypoint.parent.exists():
                raise ValueError("Passed a non-existing path!")
            self._search_pattern = self._entrypoint.name
            self._entrypoint = self._entrypoint.parent.absolute()
        else:
            self._entrypoint = self._entrypoint.absolute()
            self._search_pattern = "*"

    @property
    def entrypoint(self) -> pathlib.Path:
        return self._entrypoint

    @property
    def recursive(self) -> bool:
        return self._recursive

    @property
    def search_pattern(self) -> str:
        return self._search_pattern

    @property
    def is_single_file(self) -> bool:
        return self._entrypoint.is_file()

    @property
    def paths(self) -> pathlib.Path:
        if self.is_single_file:
            yield self._entrypoint

        if self._recursive:
            list_generator = self._entrypoint.rglob(self._search_pattern)
        else:
            list_generator = self._entrypoint.glob(self._search_pattern)

        for p in sorted(list_generator):
            if p.is_file():
                yield p

    @property
    def parent_directory(self) -> pathlib.Path:
        if self._entrypoint.is_dir():
            return self._entrypoint
        else:
            return self._entrypoint.parent


def load(
    path: Union[str, pathlib.Path], recursive: bool = True,
) -> DatasetCollection:
    """
    Lazy function for loading simulation data.

    Takes a path and tries to load all possible files into a container. If
    the path is a single file, only a container fitting for this file is
    returned. If the file is not a single file but rather a directory then
    diagnostic files are searched for recursively. You can control the search
    by providing a search pattern.

    Parameters
    ----------
    ``path`` : ``str``, ``pathlib.Path``

        Path which will be used to search for diagnostic files. If the path
        does not end on a file or a directory, the string after the last
        ``/`` will be used a search pattern.

    ``recursive`` : ``bool``, default: ``True``

        Parameter for deciding if a path should be traversed recursively
        ``True`` or if only the top directory should be searched ``False``.
        The default option is ``True``.

    Returns
    -------
    ``dataobj`` : ``GridDataset``, ``DatasetCollection``

        Returns specific container based what was provided as ``path``

    Examples
    --------
        # loads all diagnostics
        >>> data = nata.load("sim_dir/MS/")
        >>> type(data)
        nata.containers.DatasetCollection
        # e1 is a directory with only one file `e1.h5`
        >>> data = nata.load("sim_dir/MS/FLD/e1")
        >>> type(data)
        nata.containers.GridDataset
        # load data recursively by matching
        >>> data = nata.load("sim_dir/MS/e1*.h5")
    """
    filelist = _FileList(path, recursive=recursive)
    collection = DatasetCollection(root_path=filelist.parent_directory)

    for p in filelist.paths:
        collection.append(p)

    # TODO: add here a possibility to reduce the data objects fully
    #   - should return on Dataset object if only one object is present
    return collection


def activate_logging(loggin_level: str = "info"):
    import logging

    if loggin_level == "notset":
        level = logging.NOTSET
    elif loggin_level == "debug":
        level = logging.DEBUG
    elif loggin_level == "info":
        level = logging.INFO
    elif loggin_level == "warning":
        level = logging.WARNING
    elif loggin_level == "error":
        level = logging.ERROR
    elif loggin_level == "critical":
        level = logging.CRITICAL
    else:
        raise ValueError(
            "Invalid loggin level provided! "
            + "Allowed are 'notset', 'debug', 'info', 'warning', 'error', "
            + "and 'critical'!"
        )

    logging.basicConfig(format="%(levelname)s :: %(message)s", level=level)



# -*- coding: utf-8 -*-
def filter_style(cls, style: dict = dict(),) -> dict:

    style_sel = cls.style_attrs()
    # build filtered style attributes
    style_flt = dict()
    for attrib in style_sel:
        try:
            # TODO: make this use defaults - most likely it can be None @fabio
            #       - if so, try and except can be removed
            #       - else KeyError should be excepted
            prop = style.pop(attrib)
            style_flt[attrib] = prop
        except:  # noqa: E722
            continue

    return style_flt



# -*- coding: utf-8 -*-
from functools import wraps
from inspect import getfullargspec
from inspect import isclass
from inspect import isfunction
from typing import Optional


def _parse_register_args(callable_or_container, **kwargs):
    args = {"container": None, "name": None, "function": None}
    args.update(
        {key: value for key, value in kwargs.items() if args.get(key) != value}
    )

    if isfunction(callable_or_container):
        arg_spec = getfullargspec(callable_or_container)
        args["function"] = callable_or_container

        if arg_spec.args and (arg_spec.args[0] in arg_spec.annotations):
            args["container"] = arg_spec.annotations[arg_spec.args[0]]

    if isclass(callable_or_container):
        args["container"] = callable_or_container

    if args["container"] is not None:
        return args

    raise ValueError("No container to register plugin")


def register_container_plugin(
    callable_or_container=None, container=None, name: Optional[str] = None
):
    """Decorator for registering a plugin for a container."""
    arguments = _parse_register_args(
        callable_or_container, container=container, name=name
    )

    container = arguments.get("container")
    function_name = arguments.get("name")
    function = arguments.get("function")

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        container.register_plugin(
            function_name if function_name else func.__name__, wrapper
        )
        return func

    if function:
        return decorator(function)

    return decorator


# TODO: following example breaks currently -- bug OR feature :)
# @register_container_plugin(name="not_so_different")
# def really_some_different_function(dataset: nata.GridDataset):
#     print(f"inside `really_some_different_function` -> {dataset.name}")



# -*- coding: utf-8 -*-
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from .types import AxisType
from .types import GridAxisType
from .types import is_basic_indexing
from .utils.formatting import make_identifiable


def _log_axis(
    min_: Union[float, np.ndarray], max_: Union[float, np.ndarray], points: int
) -> np.ndarray:
    """Generates logarithmically spaced axis/array.

    Returns always an array with the shape (points,) + np.shape(min/max) and
    a floating dtype.
    """
    if np.issubdtype(
        # min_
        type(min_) if not isinstance(min_, np.ndarray) else min_.dtype,
        np.floating,
    ) or np.issubdtype(
        # max_
        type(max_) if not isinstance(max_, np.ndarray) else max_.dtype,
        np.floating,
    ):
        dtype = None
    else:
        dtype = float

    return np.logspace(np.log10(min_), np.log10(max_), points, dtype=dtype)


def _lin_axis(
    min_: Union[float, np.ndarray], max_: Union[float, np.ndarray], points: int
) -> np.ndarray:
    """Generates linearly spaced axis/array.

    Returns always an array with the shape (points,) + np.shape(min/max) and
    a floating dtype.
    """
    if np.issubdtype(
        # min_
        type(min_) if not isinstance(min_, np.ndarray) else min_.dtype,
        np.floating,
    ) or np.issubdtype(
        # max_
        type(max_) if not isinstance(max_, np.ndarray) else max_.dtype,
        np.floating,
    ):
        dtype = None
    else:
        dtype = float

    return np.linspace(min_, max_, points, dtype=dtype)


class Axis:
    def __init__(
        self,
        data: np.ndarray,
        *,
        name: str = "unnamed",
        label: str = "",
        unit: str = "",
    ):
        if not isinstance(data, np.ndarray):
            data = np.asanyarray(data)

        self._data = data if data.ndim > 0 else data[np.newaxis]

        name = make_identifiable(name)
        self._name = name if name else "unnamed"
        self._label = label
        self._unit = unit

    def __repr__(self) -> str:
        repr_ = f"{self.__class__.__name__}("
        repr_ += f"name='{self.name}', "
        repr_ += f"label='{self.label}', "
        repr_ += f"unit='{self.unit}', "
        repr_ += f"axis_dim={self.axis_dim}, "
        repr_ += f"len={len(self)}"
        repr_ += ")"

        return repr_

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> "Axis":
        for d in self._data:
            yield self.__class__(
                d[np.newaxis], name=self.name, label=self.label, unit=self.unit,
            )

    def __getitem__(
        self, key: Union[int, slice, Tuple[Union[int, slice]]]
    ) -> "Axis":
        if not is_basic_indexing(key):
            raise IndexError("Only basic indexing is supported!")

        key = np.index_exp[key]
        requires_new_axis = False

        # > determine if axis extension is required
        # 1st index (temporal slicing) not hidden if ndim == axis_dim + 1
        # or alternatively -> check len of the axis -> number of temporal slices
        if len(self) != 1:
            # revert dimensionality reduction
            if isinstance(key[0], int):
                requires_new_axis = True
        else:
            requires_new_axis = True

        data = self.data[key]

        if requires_new_axis:
            data = data[np.newaxis]

        return self.__class__(
            data, name=self.name, label=self.label, unit=self.unit,
        )

    def __setitem__(
        self, key: Union[int, slice, Tuple[Union[int, slice]]], value: Any
    ) -> None:
        self.data[key] = value

    def __array__(self, dtype: Optional[np.dtype] = None) -> np.ndarray:
        data = self._data.astype(dtype) if dtype else self._data
        return np.squeeze(data, axis=0) if len(self) == 1 else data

    @property
    def data(self) -> np.ndarray:
        return np.asanyarray(self)

    @data.setter
    def data(self, value: Union[np.ndarray, Any]) -> None:
        new = np.broadcast_to(value, self.shape, subok=True)
        if len(self) == 1:
            self._data = np.array(new, subok=True)[np.newaxis]
        else:
            self._data = np.array(new, subok=True)

    @property
    def axis_dim(self):
        return self._data.ndim - 1

    @property
    def shape(self):
        return self._data.shape[1:] if len(self) == 1 else self._data.shape

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def ndim(self):
        return (self._data.ndim - 1) if len(self) == 1 else self._data.ndim

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        parsed_value = make_identifiable(str(value))
        if not parsed_value:
            raise ValueError(
                "Invalid name provided! Has to be able to be valid code"
            )
        self._name = parsed_value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        value = str(value)
        self._label = value

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        value = str(value)
        self._unit = value

    def equivalent(self, other: Union[Any, AxisType]) -> bool:
        if not isinstance(other, self.__class__):
            return False

        if self.axis_dim != other.axis_dim:
            return False

        if self.name != other.name:
            return False

        if self.label != other.label:
            return False

        if self.unit != other.unit:
            return False

        return True

    def append(self, other: "Axis") -> "Axis":
        if not isinstance(other, self.__class__):
            raise TypeError(f"Can not append '{other}' to '{self}'")

        if not self.equivalent(other):
            raise ValueError(
                f"Mismatch in attributes between '{self}' and '{other}'"
            )

        selfdata = (
            self.data[np.newaxis] if self.ndim == self.axis_dim else self.data
        )

        otherdata = (
            other.data[np.newaxis]
            if other.ndim == other.axis_dim
            else other.data
        )

        self._data = np.append(selfdata, otherdata, axis=0)


_ignored_if_data = object()


class GridAxis(Axis):
    _supported_axis_types: Tuple[str, ...] = (
        "lin",
        "linear",
        "log",
        "logarithmic",
        "custom",
    )

    def __init__(
        self,
        data: np.ndarray,
        *,
        axis_type: str = "linear",
        name: str = "unnamed",
        label: str = "",
        unit: str = "",
    ) -> None:
        if axis_type not in self._supported_axis_types:
            raise ValueError(
                f"'{axis_type}' is not supported for axis_type! "
                + f"It has to by one of {self._supported_axis_types}"
            )

        super().__init__(data, name=name, label=label, unit=unit)
        self._axis_type = axis_type

    def __iter__(self) -> "GridAxis":
        for d in self._data:
            yield self.__class__(
                d[np.newaxis],
                name=self.name,
                label=self.label,
                unit=self.unit,
                axis_type=self.axis_type,
            )

    def __getitem__(
        self, key: Union[int, slice, Tuple[Union[int, slice]]]
    ) -> "GridAxis":
        if not is_basic_indexing(key):
            raise IndexError("Only basic indexing is supported!")

        key = np.index_exp[key]
        requires_new_axis = False

        # first index corresponds to temporal slicing if ndim == axis_dim + 1
        # or alternatively -> check len of the axis -> number of temporal slices
        if len(self) != 1:
            # revert dimensionality reduction
            if isinstance(key[0], int):
                requires_new_axis = True
        else:
            requires_new_axis = True

        return self.__class__(
            self.data[key][np.newaxis] if requires_new_axis else self.data[key],
            name=self.name,
            label=self.label,
            unit=self.unit,
            axis_type=self.axis_type,
        )

    def __repr__(self) -> str:
        repr_ = f"{self.__class__.__name__}("
        repr_ += f"name='{self.name}', "
        repr_ += f"label='{self.label}', "
        repr_ += f"unit='{self.unit}', "
        repr_ += f"axis_type={self.axis_type}, "
        repr_ += f"axis_dim={self.axis_dim}, "
        repr_ += f"len={len(self)}"
        repr_ += ")"

        return repr_

    @property
    def axis_type(self) -> str:
        return self._axis_type

    @axis_type.setter
    def axis_type(self, value: str) -> None:
        value = str(value)
        if value not in self._supported_axis_types:
            raise ValueError(
                f"'{value}' is not supported for axis_type! "
                + f"It has to by one of {self._supported_axis_types}"
            )
        self._axis_type = value

    @classmethod
    def from_limits(
        cls,
        min_value: Union[np.ndarray, int, float],
        max_value: Union[np.ndarray, int, float],
        cells: int,
        *,
        axis_type: str = "linear",
        name: str = "unnamed",
        label: str = "",
        unit: str = "",
    ) -> "GridAxis":
        if axis_type in ("lin", "linear"):
            axis: np.ndarray = _lin_axis(min_value, max_value, cells)
        elif axis_type in ("log", "logarithmic"):
            axis: np.ndarray = _log_axis(min_value, max_value, cells)
        else:
            raise ValueError(
                "Invalid axis type provided. "
                + "Only 'lin', 'linear', 'log', and 'logarithmic' "
                + "are supported!"
            )

        if axis.ndim == 1:
            axis = axis[np.newaxis]

        axis = cls(axis, name=name, label=label, unit=unit)
        axis._axis_type = axis_type
        return axis

    def equivalent(self, other: Union[Any, GridAxisType]) -> bool:
        if not super().equivalent(other):
            return False

        if self.axis_type != other.axis_type:
            return False

        return True



# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from nata.containers import GridDataset
from nata.containers import ParticleDataset
from nata.utils.cached_property import cached_property
from nata.utils.container_tools import register_backend
from nata.utils.zdf import info
from nata.utils.zdf import read


@register_backend(GridDataset)
class Osiris_zdf_GridFile:
    name = "osiris_dev_grid_zdf"
    location: Optional[Path] = None

    def __init__(self, location: Union[str, Path]) -> None:
        self.location = (
            location if isinstance(location, Path) else Path(location)
        )

    @staticmethod
    def is_valid_backend(file_path: Union[Path, str]) -> bool:
        if isinstance(file_path, str):
            file_path = Path(file_path)

        if not isinstance(file_path, Path):
            return False

        if not file_path.is_file():
            return False

        if not file_path.suffix == ".zdf":
            return False

        z_info = info(str(file_path))
        if hasattr(z_info, "type"):
            if z_info.type == "grid":
                return True

        return False

    @cached_property
    def _dset_name(self) -> str:
        logging.info(f"Accessing '{self.location}' for '_dset_name'")
        z_info = info(str(self.location))
        label = z_info.grid.label
        return z_info.grid.name or self.clean(label)

    def get_data(self=None, indexing=None):
        # TODO: apply indexing here
        (z_data, z_info) = read(str(self.location))
        return z_data.transpose()

    @cached_property
    def dataset_name(self) -> str:
        logging.info(f"Accessing '{self.location}' for 'dataset_name'")
        z_info = info(str(self.location))
        label = z_info.grid.label
        return z_info.grid.name or self.clean(label)

    @cached_property
    def dataset_label(self) -> str:
        logging.info(f"Accessing '{self.location}' for 'dataset_label'")
        z_info = info(str(self.location))
        return z_info.grid.label

    @cached_property
    def ndim(self):
        logging.info(f"Accessing '{self.location}' for 'ndim'")
        z_info = info(str(self.location))
        return z_info.grid.ndims

    @cached_property
    def shape(self):
        logging.info(f"Accessing '{self.location}' for 'shape'")
        z_info = info(str(self.location))
        nx = []
        for n in z_info.grid.nx:
            nx.append(n.item())
        return tuple(nx)

    @cached_property
    def dtype(self):
        logging.info(f"Accessing '{self.location}' for 'dtype'")
        (z_data, z_info) = read(str(self.location))
        return z_data.dtype

    @cached_property
    def dataset_unit(self):
        logging.info(f"Accessing '{self.location}' for 'dataset_unit'")
        z_info = info(str(self.location))
        return z_info.grid.units

    @cached_property
    def axes_min(self):
        logging.info(f"Accessing '{self.location}' for 'axes_min'")
        min_values = []
        z_info = info(str(self.location))
        for axis in z_info.grid.axis:
            min_values.append(axis.min)
        return np.array(min_values)

    @cached_property
    def axes_max(self):
        logging.info(f"Accessing '{self.location}' for 'axes_max'")
        max_values = []
        z_info = info(str(self.location))
        for axis in z_info.grid.axis:
            max_values.append(axis.max)
        return np.array(max_values)

    @cached_property
    def axes_names(self):
        logging.info(f"Accessing '{self.location}' for 'axes_names'")
        names = []
        z_info = info(str(self.location))
        for axis in z_info.grid.axis:
            label = axis.label
            names.append(axis.name or self.clean(label))
        return np.array(names)

    @cached_property
    def axes_labels(self):
        logging.info(f"Accessing '{self.location}' for 'axes_labels'")
        long_names = []
        z_info = info(str(self.location))
        for axis in z_info.grid.axis:
            long_names.append(axis.label)
        return np.array(long_names)

    @cached_property
    def axes_units(self):
        logging.info(f"Accessing '{self.location}' for 'axes_units'")
        units = []
        z_info = info(str(self.location))
        for axis in z_info.grid.axis:
            units.append(axis.units)
        return np.array(units)

    @cached_property
    def iteration(self):
        logging.info(f"Accessing '{self.location}' for 'iteration'")
        z_info = info(str(self.location))
        return z_info.iteration.n

    @cached_property
    def time_step(self):
        logging.info(f"Accessing '{self.location}' for 'time_step'")
        z_info = info(str(self.location))
        return z_info.iteration.t

    @cached_property
    def time_unit(self):
        logging.info(f"Accessing '{self.location}' for 'time_unit'")
        z_info = info(str(self.location))
        return z_info.iteration.tunits

    def clean(self, name):
        return (
            name.lower()
            .replace("_", "")
            .replace("^", "")
            .replace("{", "")
            .replace("}", "")
            .replace("\\", "")
            .replace(" ", "_")
        )


@register_backend(ParticleDataset)
class Osiris_zdf_ParticleFile:
    name = "osiris_dev_particles_zdf"
    location: Optional[Path] = None

    def __init__(self, location=Union[str, Path]) -> None:
        self.location = (
            location if isinstance(location, Path) else Path(location)
        )

    @staticmethod
    def is_valid_backend(file_path: Path) -> bool:
        if not file_path.is_file():
            return False

        if not file_path.suffix == ".zdf":
            return False

        z_info = info(str(file_path))
        if hasattr(z_info, "type"):
            if z_info.type == "particles":
                return True

        return False

    @cached_property
    def dataset_name(self) -> str:
        logging.info(f"Accessing '{self.location}' for 'dataset_name'")
        z_info = info(str(self.location))
        return z_info.particles.name

    @cached_property
    def num_particles(self) -> int:
        logging.info(f"Accessing '{self.location}' for 'num_particles'")
        z_info = info(str(self.location))
        return z_info.particles.nparts

    def get_data(self, indexing=None, fields=None) -> np.ndarray:
        logging.info(f"Reading data in '{self.location}'")
        (z_data, z_info) = read(str(self.location))
        if fields is None:
            # create a structured array
            dset = np.empty(self.num_particles, dtype=self.dtype)

            # fill the array
            for quant in self.quantity_names:
                dset[quant] = z_data[quant]
        else:
            if indexing is None:
                dset = z_data[fields][:]
            else:
                dset = z_data[fields][indexing]

        return dset

    @cached_property
    def quantity_names(self) -> List[str]:
        logging.info(f"Accessing '{self.location}' for 'quantity_names'")
        z_info = info(str(self.location))
        quantities = []
        for key in z_info.particles.quants:
            if key == "tag":
                continue
            quantities.append(key)

        return quantities

    @cached_property
    def quantity_labels(self) -> List[str]:
        logging.info(f"Accessing '{self.location}' for 'quantity_labels'")
        z_info = info(str(self.location))
        names = []
        for quant in self.quantity_names:
            names.append(z_info.particles.qlabels[quant])
        return names

    @cached_property
    def quantity_units(self) -> List[str]:
        logging.info(f"Accessing '{self.location}' for 'quantity_units'")
        z_info = info(str(self.location))
        units = []
        for quant in self.quantity_names:
            units.append(z_info.particles.qunits[quant])

        return units

    @cached_property
    def dtype(self) -> np.dtype:
        logging.info(f"Accessing '{self.location}' for 'dtype'")
        (z_data, z_info) = read(str(self.location))
        fields = []
        for quant in self.quantity_names:
            fields.append((quant, z_data[quant].dtype))

        return np.dtype(fields)

    @cached_property
    def iteration(self) -> int:
        logging.info(f"Accessing '{self.location}' for 'iteration'")
        z_info = info(str(self.location))
        return z_info.iteration.n

    @cached_property
    def time_step(self) -> float:
        logging.info(f"Accessing '{self.location}' for 'time_step'")
        z_info = info(str(self.location))
        return z_info.iteration.t

    @cached_property
    def time_unit(self) -> str:
        logging.info(f"Accessing '{self.location}' for 'time_unit'")
        z_info = info(str(self.location))
        return z_info.iteration.tunits



# -*- coding: utf-8 -*-
def inside_ipython():
    try:
        get_ipython()
        inside = True
    except NameError:
        inside = False

    return inside


def inside_notebook():
    if inside_ipython():
        shell = get_ipython().__class__.__name__  # noqa: F821
        if shell == "ZMQInteractiveShell":
            return True

    return False


def inside_ipython_terminal():
    if inside_ipython():
        shell = get_ipython().__class__.__name__  # noqa: F821
        if shell == "TerminalInteractiveShell":
            return True

    return False



# -*- coding: utf-8 -*-
from logging import info
from pathlib import Path
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import h5py as h5
import numpy as np

import ndindex as ndx
from nata.containers import GridDataset
from nata.containers import ParticleDataset
from nata.types import BasicIndex
from nata.types import BasicIndexing
from nata.types import FileLocation
from nata.utils.backends import sort_particle_quantities
from nata.utils.container_tools import register_backend


@register_backend(GridDataset)
class Osiris_Hdf5_GridFile:
    name = "osiris_4.4.4_grid_hdf5"
    location: Optional[Path] = None

    def __init__(self, location: Union[str, Path]) -> None:
        self.location = (
            location if isinstance(location, Path) else Path(location)
        )

        info(f"Obtaining backend props for '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            short_name = fp.attrs["NAME"].astype(str)[0]
            if short_name in fp:
                self._dset_name = short_name
            else:
                name_ = short_name.split()[-1]
                self._dset_name = name_.replace("_", "")

            self._dataset_name = fp.attrs["NAME"].astype(str)[0]
            self._dataset_label = (
                fp[self._dset_name].attrs["LONG_NAME"].astype(str)[0]
            )
            self._ndim = fp[self._dset_name].ndim
            self._shape = fp[self._dset_name].shape[::-1]
            self._dtype = fp[self._dset_name].dtype
            self._units = fp[self._dset_name].attrs["UNITS"].astype(str)[0]

            min_, max_ = [], []
            axes_names, axes_labels, axes_units = [], [], []
            for ax in fp["AXIS"]:
                min_.append(fp["AXIS/" + ax][0])
                max_.append(fp["AXIS/" + ax][1])
                axes_names.append(fp["AXIS/" + ax].attrs["NAME"].astype(str)[0])
                axes_labels.append(
                    fp["AXIS/" + ax].attrs["LONG_NAME"].astype(str)[0]
                )
                axes_units.append(
                    fp["AXIS/" + ax].attrs["UNITS"].astype(str)[0]
                )

            self._min = np.array(min_)
            self._max = np.array(max_)
            self._axes_names = np.array(axes_names)
            self._axes_labels = np.array(axes_labels)
            self._axes_units = np.array(axes_units)

            self._iteration = fp.attrs["ITER"].astype(int)[0]
            self._time_step = fp.attrs["TIME"][0]
            self._time_unit = fp.attrs["TIME UNITS"].astype(str)[0]

    @staticmethod
    def is_valid_backend(path: FileLocation) -> bool:
        if isinstance(path, str):
            path = Path(path)

        if (
            not isinstance(path, Path)
            or not path.is_file()
            or not path.suffix == ".h5"
            or not h5.is_hdf5(path)
        ):
            return False

        with h5.File(path, mode="r") as f:
            if (
                ("NAME" in f.attrs)
                and ("TYPE" in f.attrs)
                and ("LABEL" not in f.attrs)
            ):
                type_ = f.attrs["TYPE"].astype(str)[0]
                # general naming
                name_ = f.attrs["NAME"].astype(str)[0]
                names = (name_,)
                # special case naming
                name_ = name_.split()[-1]
                name_ = name_.replace("_", "")

                names += (name_,)
                if (type_ == "grid") and any([name in f for name in names]):
                    return True

        return False

    def get_data(self, indexing: Optional[BasicIndexing] = None) -> np.ndarray:
        info(f"Reading data in '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            if indexing:
                dataset = fp[self._dset_name][indexing[::-1]]
            else:
                dataset = fp[self._dset_name][:]
        return dataset.transpose()

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    @property
    def dataset_label(self) -> str:
        return self._dataset_label

    @property
    def ndim(self):
        return self._ndim

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return self._dtype

    @property
    def dataset_unit(self):
        return self._units

    @property
    def axes_min(self):
        return self._min

    @property
    def axes_max(self):
        return self._max

    @property
    def axes_names(self):
        return self._axes_names

    @property
    def axes_labels(self):
        return self._axes_labels

    @property
    def axes_units(self):
        return self._axes_units

    @property
    def iteration(self):
        return self._iteration

    @property
    def time_step(self):
        return self._time_step

    @property
    def time_unit(self):
        return self._time_unit


@register_backend(GridDataset)
class Osiris_Dev_Hdf5_GridFile:
    name = "osiris_dev_grid_hdf5"
    location: Optional[Path] = None

    def __init__(self, location: FileLocation) -> None:
        self.location = (
            location if isinstance(location, Path) else Path(location)
        )

        info(f"Obtaining backend props for '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            short_name = fp.attrs["NAME"].astype(str)[0]
            if short_name in fp:
                self._dset_name = short_name

            name_ = short_name.split()[-1]
            name_ = name_.replace("_", "")
            if name_ in fp:
                self._dset_name = name_

            self._dataset_name = fp.attrs["NAME"].astype(str)[0]
            self._dataset_label = fp.attrs["LABEL"].astype(str)[0]
            self._dataset_unit = fp.attrs["UNITS"].astype(str)[0]
            self._ndim = fp[self._dset_name].ndim
            self._shape = fp[self._dset_name].shape[::-1]
            self._dtype = fp[self._dset_name].dtype

            min_, max_ = [], []
            axes_names, axes_labels, axes_units = [], [], []
            for ax in fp["AXIS"]:
                min_.append(fp["AXIS/" + ax][0])
                max_.append(fp["AXIS/" + ax][1])
                axes_names.append(fp["AXIS/" + ax].attrs["NAME"].astype(str)[0])
                axes_labels.append(
                    fp["AXIS/" + ax].attrs["LONG_NAME"].astype(str)[0]
                )
                axes_units.append(
                    fp["AXIS/" + ax].attrs["UNITS"].astype(str)[0]
                )

            self._axes_min = np.array(min_)
            self._axes_max = np.array(max_)
            self._axes_names = np.array(axes_names)
            self._axes_labels = np.array(axes_labels)
            self._axes_units = np.array(axes_units)

            self._iteration = fp.attrs["ITER"].astype(int)[0]
            self._time_step = fp.attrs["TIME"][0]
            self._time_unit = fp.attrs["TIME UNITS"].astype(str)[0]

    @staticmethod
    def is_valid_backend(path: FileLocation) -> bool:
        if isinstance(path, str):
            path = Path(path)

        if (
            not isinstance(path, Path)
            or not path.is_file()
            or not path.suffix == ".h5"
            or not h5.is_hdf5(path)
        ):
            return False

        with h5.File(path, mode="r") as f:
            if (
                ("NAME" in f.attrs)
                and ("TYPE" in f.attrs)
                and ("LABEL" in f.attrs)
            ):
                type_ = f.attrs["TYPE"].astype(str)[0]
                # general naming
                name_ = f.attrs["NAME"].astype(str)[0]
                names = (name_,)
                # special case naming
                name_ = name_.split()[-1]
                name_ = name_.replace("_", "")

                names += (name_,)
                if (type_ == "grid") and any([name in f for name in names]):
                    return True

        return False

    def get_data(self, indexing: Optional[BasicIndexing] = None) -> np.ndarray:
        info(f"Reading data in '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            if indexing:
                dataset = fp[self._dset_name][indexing[::-1]]
            else:
                dataset = fp[self._dset_name][:]
        return dataset.transpose()

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    @property
    def dataset_label(self) -> str:
        return self._dataset_label

    @property
    def dataset_unit(self):
        return self._dataset_unit

    @property
    def ndim(self) -> int:
        return self._ndim

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return self._dtype

    @property
    def axes_min(self):
        return self._axes_min

    @property
    def axes_max(self):
        return self._axes_max

    @property
    def axes_names(self):
        return self._axes_names

    @property
    def axes_labels(self):
        return self._axes_labels

    @property
    def axes_units(self):
        return self._axes_units

    @property
    def iteration(self):
        return self._iteration

    @property
    def time_step(self):
        return self._time_step

    @property
    def time_unit(self):
        return self._time_unit


@register_backend(ParticleDataset)
class Osiris_Hdf5_ParticleFile:
    name = "osiris_4.4.4_particles_hdf5"
    location: Optional[Path] = None

    def __init__(self, location: FileLocation) -> None:
        self.location = (
            location if isinstance(location, Path) else Path(location)
        )

        info(f"Obtaining backend props for '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            self._dataset_name = fp.attrs["NAME"].astype(str)[0]
            self._num_particles = fp["q"].shape[0] if fp["q"].shape else 0

            # find first all quantaties - with their names
            names = []
            for key, item in fp.items():
                if key == "tag":
                    continue

                if isinstance(item, h5.Dataset):
                    names.append(key)

            self._quantity_names = sort_particle_quantities(names, ("x", "p"))

            # iterate over all the sorted names and fill other props
            labels = []
            units = []
            dtype = []
            for name in self._quantity_names:
                labels.append(fp[name].attrs["LONG_NAME"].astype(str)[0])
                units.append(fp[name].attrs["UNITS"].astype(str)[0])
                dtype.append((name, fp[name].dtype))

            self._quantity_labels = labels
            self._quantity_units = units
            self._dtype = np.dtype(dtype)

            # temporal information
            self._iteration = fp.attrs["ITER"][0]
            self._time_step = fp.attrs["TIME"][0]
            self._time_unit = fp.attrs["TIME UNITS"].astype(str)[0]

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    @property
    def num_particles(self) -> int:
        return self._num_particles

    @property
    def quantity_names(self) -> List[str]:
        return self._quantity_names

    @property
    def quantity_labels(self) -> List[str]:
        return self._quantity_labels

    @property
    def quantity_units(self) -> List[str]:
        return self._quantity_units

    @property
    def dtype(self) -> np.dtype:
        return self._dtype

    @property
    def iteration(self) -> int:
        return self._iteration

    @property
    def time_step(self) -> float:
        return self._time_step

    @property
    def time_unit(self) -> str:
        return self._time_unit

    @staticmethod
    def is_valid_backend(path: FileLocation) -> bool:
        if isinstance(path, str):
            path = Path(path)

        if (
            not isinstance(path, Path)
            or not path.is_file()
            or not path.suffix == ".h5"
            or not h5.is_hdf5(path)
        ):
            return False

        with h5.File(path, mode="r") as f:
            if ("TYPE" in f.attrs) and ("LABELS" not in f.attrs):
                if f.attrs["TYPE"].astype(str)[0] == "particles":
                    return True

        return False

    def get_data(
        self,
        indexing: Optional[BasicIndex] = None,
        fields: Optional[Union[str, Sequence[str]]] = None,
    ) -> np.ndarray:
        info(f"Reading data in '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            index = (
                ndx.Slice(None) if indexing is None else ndx.ndindex(indexing)
            )
            index = index.reduce((self.num_particles,))
            dtype = self.dtype if fields is None else self.dtype[fields]

            if dtype.fields:
                # create array as source to store quantities
                dset = np.empty(len(index), dtype=dtype)

                for field in dtype.fields:
                    dset[field][:] = fp[field][index.raw]
            else:
                # only one field element -> string passed
                dset = fp[fields][index.raw]

        return dset


@register_backend(ParticleDataset)
class Osiris_Dev_Hdf5_ParticleFile:
    name = "osiris_dev_particles_hdf5"
    location: Optional[Path] = None

    def __init__(self, location=Union[str, Path]) -> None:
        self.location = (
            location if isinstance(location, Path) else Path(location)
        )

        info(f"Obtaining backend props for '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            self._dataset_name = fp.attrs["NAME"].astype(str)[0]
            self._num_particles = fp["q"].shape[0] if fp["q"].shape else 0

            # quantaties
            unordered_names = list(fp.attrs["QUANTS"].astype(str))
            unordered_labels = list(fp.attrs["LABELS"].astype(str))
            unordered_units = list(fp.attrs["UNITS"].astype(str))

            clean_names = [name for name in unordered_names if name != "tag"]
            clean_names = sort_particle_quantities(clean_names, ("x", "p"))

            clean_labels = [
                unordered_labels[unordered_names.index(s)] for s in clean_names
            ]
            clean_units = [
                unordered_units[unordered_names.index(s)] for s in clean_names
            ]
            dtype = [(name, fp[name].dtype) for name in clean_names]

            self._quantity_names = clean_names
            self._quantity_labels = clean_labels
            self._quantity_units = clean_units
            self._dtype = np.dtype(dtype)

            # temporal information
            self._iteration = fp.attrs["ITER"][0]
            self._time_step = fp.attrs["TIME"][0]
            self._time_unit = fp.attrs["TIME UNITS"].astype(str)[0]

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    @property
    def num_particles(self) -> int:
        return self._num_particles

    @property
    def quantity_names(self) -> List[str]:
        return self._quantity_names

    @property
    def quantity_labels(self) -> List[str]:
        return self._quantity_labels

    @property
    def quantity_units(self) -> List[str]:
        return self._quantity_units

    @property
    def dtype(self) -> np.dtype:
        return self._dtype

    @property
    def iteration(self) -> int:
        return self._iteration

    @property
    def time_step(self) -> float:
        return self._time_step

    @property
    def time_unit(self) -> str:
        return self._time_unit

    @staticmethod
    def is_valid_backend(path: FileLocation) -> bool:
        if isinstance(path, str):
            path = Path(path)

        if (
            not isinstance(path, Path)
            or not path.is_file()
            or not path.suffix == ".h5"
            or not h5.is_hdf5(path)
        ):
            return False

        with h5.File(path, mode="r") as f:
            if ("TYPE" in f.attrs) and ("LABELS" in f.attrs):
                if f.attrs["TYPE"].astype(str)[0] == "particles":
                    return True

        return False

    def get_data(
        self,
        indexing: Optional[BasicIndex] = None,
        fields: Optional[Union[str, Sequence[str]]] = None,
    ) -> np.ndarray:
        info(f"Reading data in '{self.location}'")
        with h5.File(self.location, mode="r") as fp:
            index = (
                ndx.Slice(None) if indexing is None else ndx.ndindex(indexing)
            )
            index = index.reduce((self.num_particles,))
            dtype = self.dtype if fields is None else self.dtype[fields]

            if dtype.fields:
                # create array as source to store quantities
                dset = np.empty(len(index), dtype=dtype)

                for field in dtype.fields:
                    dset[field][:] = fp[field][index.raw]
            else:
                # only one field element -> string passed
                dset = fp[fields][index.raw]

        return dset



# -*- coding: utf-8 -*-
from warnings import warn

from nata.types import BackendType
from nata.types import DatasetType


def register_backend(container: DatasetType):
    """Decorater for registering a backend for a Datset"""
    if not isinstance(container, DatasetType):
        raise TypeError(f"Requires container of type '{DatasetType}'")

    def add_backend_to_container(backend: BackendType):
        if container.is_valid_backend(backend):
            container.add_backend(backend)
        else:
            warn(
                f"{backend} is an invalid backend for '{container}'. "
                + "Skipping backend registration!"
            )
        return backend

    return add_backend_to_container



# -*- coding: utf-8 -*-
from typing import Union

from .base import BasePlot

from .line import LinePlot
from .color import ColorPlot
from .scatter import ScatterPlot

PlotTypes = Union[LinePlot]

# default plots for data types
DefaultGridPlotTypes = {1: LinePlot, 2: ColorPlot}
DefaultParticlePlotType = ScatterPlot



# -*- coding: utf-8 -*-
from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from IPython.display import display
from ipywidgets import Layout
from ipywidgets import widgets
from nata.containers import GridDataset
from nata.plots.axes import Axes
from nata.plots.figure import Figure


@dataclass
class PlotPlan:
    dataset: GridDataset = object()
    style: Optional[Dict] = field(default_factory=dict)


@dataclass
class AxesPlan:
    plots: List[PlotPlan] = field(default_factory=list)
    axes: Optional[Axes] = object()
    style: Optional[Dict] = field(default_factory=dict)

    @property
    def datasets(self) -> list:
        ds = []
        for p in self.plots:
            ds.append(p.dataset)
        return ds


@dataclass
class FigurePlan:
    axes: List[AxesPlan] = field(default_factory=list)
    fig: Optional[Figure] = object()
    style: Optional[Dict] = field(default_factory=dict)

    @property
    def datasets(self) -> list:
        ds = []
        for a in self.axes:
            for d in a.datasets:
                ds.append(d)
        return ds

    def __getitem__(self, key: Union[int, slice]):
        axes = []
        for a in self.axes:
            plots = []
            for p in a.plots:
                plots.append(PlotPlan(dataset=p.dataset[key], style=p.style,))
            axes.append(AxesPlan(plots=plots, axes=a.axes, style=a.style,))

        return self.__class__(axes=axes, fig=self.fig, style=self.style,)

    def build(self) -> Figure:
        fig = self.fig if self.fig is not None else Figure(**self.style)

        for a in self.axes:
            axes = a.axes if a.axes is not None else fig.add_axes(style=a.style)

            for p in a.plots:
                plot_type = p.dataset.plot_type()
                plot_data = p.dataset.plot_data()
                axes.add_plot(
                    plot_type=plot_type, data=plot_data, style=p.style
                )

            axes.update()

        fig.close()

        return fig

    def build_interactive(self, n: int = 0):
        # get reference dataset
        d_ref = self.datasets[0]

        time = np.array(d_ref.axes["time"])
        iteration = np.array(d_ref.axes["iteration"])

        dropdown = widgets.Dropdown(
            options=["File Index", "Iteration", "Time"],
            value="File Index",
            disabled=False,
            layout=Layout(max_width="100px"),
            continuous_update=False,
        )

        slider = widgets.SelectionSlider(
            options=[f"{i}" for i in np.arange(len(iteration))],
            value=f"{n}",
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
        )

        def dropdown_change(change):

            if change.old in ["Time", "Iteration"]:
                options = np.array(slider.options).astype(np.float)
                n = np.argmax(options >= float(slider.value)).item()
            else:
                n = int(slider.value)

            with out.hold_trait_notifications():
                if change.new == "Time":
                    slider.options = [f"{i:.2f}" for i in time]
                    slider.value = f"{time[n]:.2f}"

                elif change.new == "Iteration":
                    slider.options = [f"{i:d}" for i in iteration]
                    slider.value = f"{iteration[n]:d}"
                else:
                    slider.options = [
                        f"{i:n}" for i in np.arange(len(iteration))
                    ]
                    slider.value = f"{n:d}"

        dropdown.observe(dropdown_change, names=["value"], type="change")

        ui = widgets.HBox([dropdown, slider])

        def update_figure(sel):
            if dropdown.value == "Time":
                n = np.argmax(time >= float(sel)).item()
            elif dropdown.value == "Iteration":
                n = np.argmax(iteration >= int(sel)).item()
            else:
                n = int(sel)

            f_a = []

            for a in self.axes:
                a_p = []
                for p in a.plots:
                    a_p.append(PlotPlan(dataset=p.dataset[n], style=p.style))

                f_a.append(AxesPlan(axes=None, plots=a_p, style=a.style))

            f_p = FigurePlan(fig=None, axes=f_a, style=self.style)

            # build figure
            fig = f_p.build()

            return fig.show()

        out = widgets.interactive_output(update_figure, {"sel": slider})

        display(ui, out)



# -*- coding: utf-8 -*-
from typing import Tuple
from typing import Union

import numpy as np


def expand_ellipsis(key, dimensions: int) -> Tuple[Union[int, slice], ...]:
    key = np.index_exp[key]

    if all(k is not Ellipsis for k in key):
        return key

    if key.count(Ellipsis) > 1:
        raise KeyError("Only one Ellipse '...' is allowed!")

    key = list(key)
    cleaned_key = list(
        filter(lambda k: (k is not None) and (k is not Ellipsis), key)
    )

    ellipsis_expanded = [
        slice(None) for _ in range(dimensions - len(cleaned_key))
    ]

    index_of_ellipse = key.index(Ellipsis)
    expanded_key = (
        key[:index_of_ellipse]
        + ellipsis_expanded
        + key[(index_of_ellipse + 1) :]
    )

    return tuple(expanded_key)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nata',
 'nata.backends',
 'nata.backends.osiris',
 'nata.plots',
 'nata.plots.types',
 'nata.plugins',
 'nata.plugins.grids',
 'nata.plugins.particles',
 'nata.plugins.plot',
 'nata.utils']

package_data = \
{'': ['*'], 'nata.plots': ['styles/*']}

install_requires = \
['h5py>=2.10.0,<3.0.0',
 'ipywidgets>=7.5.1,<8.0.0',
 'matplotlib>=3.1.2,<4.0.0',
 'ndindex>=1.3.1,<2.0.0',
 'numpy>=1.18.1,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['typing_extensions>=3.7.4,<4.0.0'],
 ':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 'docs': ['sphinx>=3.0.3,<4.0.0',
          'sphinx-rtd-theme>=0.4.3,<0.5.0',
          'recommonmark>=0.6.0,<0.7.0'],
 'wiptools': ['jupyterlab>=2.1.2,<3.0.0']}

setup_kwargs = {
    'name': 'nata',
    'version': '0.0.3',
    'description': 'Post-processing and visualization for PIC codes',
    'long_description': '<p align="center">\n     <img\n          src="https://raw.githubusercontent.com/GoLP-IST/nata/master/docs/_static/nata-logo.png"\n          alt="nata logo"\n          width=460\n     />\n</p>\n\n<p align="center">\n<a href="https://pypi.org/project/nata/"><img alt="PyPI" src="https://img.shields.io/pypi/v/nata"></a>\n<a href=\'https://nata.readthedocs.io/en/master/?badge=master\'>\n    <img src=\'https://readthedocs.org/projects/nata/badge/?version=master\' alt=\'Documentation Status\' />\n</a>\n</p>\n\n\n**Nata** is a python package for post-processing and visualizing simulation\noutput for particle-in-cell codes. It utilizes the numpy interface to provide\na simple way to read, manipulate, and represent simulation output.\n\n## Installing nata\n\nNata is available on PyPI. You can install it by running the following\ncommand inside your terminal\n\n```shell\npip install nata\n```\n\nIt can be used inside an IPython shell or [jupyter notebook](https://jupyter.org/) together\nwith [ipywidgets](https://github.com/jupyter-widgets/ipywidgets). Hence, you\nmight need to run after the installation\n\n```shell\n# can be skipped for notebook version 5.3 and above\njupyter nbextension enable --py --sys-prefix widgetsnbextension\n```\n\nand if you want to use it inside JupyterLab (note that this requires nodejs\nto be installed)\n\n```bash\njupyter labextension install @jupyter-widgets/jupyterlab-manager\n```\n\nIn case of issues, please visit the [installation section of ipywidgets](https://github.com/jupyter-widgets/ipywidgets/blob/master/docs/source/user_install.md)\nfor further details.\n\n## Contributing to nata\n\nAny type of contribution to nata is appreciated. If you have any issues,\nplease report them [by adding an issue on GitHub](https://github.com/GoLP-IST/nata/issues). But if\nyou wish to directly contribute to nata, we recommend to setup a local\ndevelopment environment. Follow the instruction below for more details.\n\n### Getting the source code\n\nThe source code is hosted on GitHub. Simply create a fork and apply your\nchanges. You can always push any changes to your local fork, but if you would\nlike to share your contribution, please create a pull request, so that it can\nbe reviewed.\n\n### Local Development Environment\n\nFor the local development environment, we use\n[poetry](https://python-poetry.org/). This allows us to deal better with\ndependencies issues and to ensure coding standards without the burden of\nmanually fixing and checking for styles. To use poetry, simply install it\nusing the latest version by running the following command in your terminal.\n\n```shell\ncurl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python\n```\n\nAfterwards, simply run\n\n```shell\npoetry install\n```\n\nwhich will install all the dependencies (including development packages) in a\nvirtual environment. If you wish to run a command inside the virtual\nenvironment, simply run it by `poetry run <your command>`, e.g. to run all\nthe test\n\n```shell\npoetry run pytest tests\n```\n\nAlternatively, you can directly use the virtual environment by running the\ncommand\n\n```shell\npoetry shell\n```\n\nwhich will spawn a shell inside the virtual environment. With this, you can\neven run jupyter outside of the source directory.\n\nIn addition, we use [pre-commit](https://pre-commit.com/) to help us keep\nconsistency in our development without any additional burden. Please use it\nas well. Inside the root directory of the repository, run the command\n\n```shell\npoetry run pre-commit install\n```\n\nwhich will create commit hooks for you and modify files keeping a consistent\nstructure.\n\n# Credits\n\n**Nata** is created and maintained by [Anton Helm](https://github.com/ahelm)\nand [Fábio Cruz](https://github.com/fabiocruz).\n\nThe development is kindly supported by the [Group for Lasers and Plasmas (GoLP)](http://epp.tecnico.ulisboa.pt/>).\n',
    'author': 'Anton Helm',
    'author_email': 'anton.helm@tecnico.ulisboa.pt',
    'maintainer': 'Anton Helm',
    'maintainer_email': 'anton.helm@tecnico.ulisboa.pt',
    'url': 'https://github.com/GoLP-IST/nata',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
# grid backends
from .hdf5 import Osiris_Hdf5_GridFile
from .hdf5 import Osiris_Dev_Hdf5_GridFile
from .zdf import Osiris_zdf_GridFile

# particle backends
from .hdf5 import Osiris_Hdf5_ParticleFile
from .hdf5 import Osiris_Dev_Hdf5_ParticleFile
from .zdf import Osiris_zdf_ParticleFile
